# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# To change this template, choose Tools | Templates
# and open the template in the editor.

import unittest

from karabo.bound import (
    INT32_ELEMENT, KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS,
    STRING_ELEMENT, Configurator, Hash, Types)
from karabo.common.states import State


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("ExampleBase", "1.0")
class ExampleBaseClass:
    def __init__(self, configuration):
        self.configuration = configuration

    @staticmethod
    def expectedParameters(expected):
        e = STRING_ELEMENT(expected)
        e.key('firstWord').displayedName("First Word").description(
            "Input for first word")
        e.assignmentOptional().defaultValue("")
        e.reconfigurable().commit()

        e = STRING_ELEMENT(expected)
        e.key("secondWord").displayedName("Second Word").description(
            "Input for first word")
        e.assignmentOptional().defaultValue("")
        e.reconfigurable().commit()

        e = INT32_ELEMENT(expected)
        e.key("multiply").displayedName("Multiply").description(
            "multiplies word")
        e.assignmentOptional().defaultValue(1)
        e.allowedStates(State.ERROR).reconfigurable().commit()

        e = STRING_ELEMENT(expected)
        e.key("composedWord").displayedName("Composed word").description(
            "The composed word")
        e.assignmentOptional().noDefaultValue()
        e.readOnly().commit()


@KARABO_CLASSINFO("ExampleDerived", "1.0")
class ExampleDerivedClass(ExampleBaseClass):
    def __init__(self, configuration):
        super().__init__(configuration)
        self.configuration = configuration

    @staticmethod
    def expectedParameters(expected):
        e = STRING_ELEMENT(expected)
        e.key('hostname').displayedName("Hostname").description(
            "Input for host name")
        e.assignmentOptional().defaultValue("")
        e.reconfigurable().commit()

        e = INT32_ELEMENT(expected)
        e.key("port").displayedName("Port").description("Input for port")
        e.assignmentOptional().defaultValue(2999)
        e.reconfigurable().commit()


@KARABO_CLASSINFO("Example", "1.0")
class ExampleClass(ExampleBaseClass):
    def __init__(self, configuration):
        super().__init__(configuration)
        self.configuration = configuration

    @staticmethod
    def expectedParameters(expected):
        e = STRING_ELEMENT(expected)
        e.key('street').displayedName("Street").description("Street")
        e.assignmentOptional().defaultValue("Albert-Einstein-Ring")
        e.reconfigurable().commit()

        e = INT32_ELEMENT(expected)
        e.key("bld").displayedName("Bld").description("Building number")
        e.assignmentOptional().defaultValue(19)
        e.reconfigurable().commit()


# ------------------------------------------------
# Inherit from a class without Karabo

class NoKaraboClass:
    def __init__(self, configuration):
        super().__init__(configuration)

    def getKaraboVersion(self):
        return 2


@KARABO_CLASSINFO("ExampleNoKarabo", "1.0")
class ExampleNoKarabo(NoKaraboClass, ExampleDerivedClass):
    def __init__(self, configuration):
        super().__init__(configuration)
        self.configuration = configuration

    @staticmethod
    def expectedParameters(expected):
        e = STRING_ELEMENT(expected)
        e.key('hostname').displayedName("Hostname").description(
            "Input for host name")
        e.assignmentOptional().defaultValue("")
        e.reconfigurable().commit()

        e = INT32_ELEMENT(expected)
        e.key("port").displayedName("Port").description("Input for port")
        e.assignmentOptional().defaultValue(2999)
        e.reconfigurable().commit()


class Decorators_TestCase(unittest.TestCase):
    def test_decorators_(self):
        try:
            # call Configurator by classid
            schema = Configurator("ExampleBase").getSchema('Example')
            self.assertIsNotNone(schema)
            # call Configurator by class
            schema = Configurator(ExampleBaseClass).getSchema('Example')
            self.assertEqual(
                schema.getValueType("firstWord"), Types.STRING,
                "expectedParameters failed -- no 'firstWord' key found")
            self.assertEqual(
                schema.getValueType("secondWord"), Types.STRING,
                "expectedParameters failed -- no 'secondWord' key found")
            self.assertEqual(
                schema.getValueType("multiply"), Types.INT32,
                "expectedParameters failed -- no 'multiply' key found")
            self.assertEqual(
                schema.getValueType("composedWord"), Types.STRING,
                "expectedParameters failed -- no 'composedWord' key found")

            # Just try the other methods
            example = Configurator('ExampleBase').create('Example', Hash())
            self.assertIsNotNone(example)
            schema_derived = Configurator(ExampleBaseClass).getSchema(
                'Example')
            self.assertIsNotNone(schema_derived)
            schema_inherit = ExampleNoKarabo.create(
                'ExampleNoKarabo', Hash()).getSchema('ExampleNoKarabo')
            self.assertFalse(schema_inherit.empty())
            self.assertTrue(schema_inherit.has('hostname'))
            self.assertTrue(schema_inherit.has('port'))

        except Exception as e:
            self.fail("test_decorators group 1: " + str(e))


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.bound import (
    HashAttributes as Attributes, HashAttributesNode as Node, Types)


def test_constructor():
    # Can contruct only empty hash attributes
    ha = Attributes()
    assert len(ha) == 0
    assert ha.empty()
    # No args are accepted
    with pytest.raises(TypeError):
        ha = Attributes('a', 1)

    # Construction of HashAttributesNode is not possible in python
    # (but exceptions differ)
    with pytest.raises(TypeError):
        _ = Node()


def test_set():
    # Set attribute
    ha = Attributes()
    # BOOL
    ha.set('attr1', False)
    assert ha.getNode('attr1').getType() == Types.BOOL
    assert not ha.get('attr1')
    # INT32
    ha.set('attr2', 12)
    assert ha.getNode('attr2').getType() == Types.INT32
    assert ha.get('attr2') == 12
    # INT64
    ha.set('attr3', 333333333333)
    assert ha.getNode('attr3').getType() == Types.INT64
    assert ha.get('attr3') == 333333333333
    # DOUBLE
    ha.set('attr4', 3.14159265359)
    assert ha.getNode('attr4').getType() == Types.DOUBLE
    assert ha.get('attr4') == 3.14159265359
    # COMPLEX_DOUBLE
    ha.set('attr5', complex(12., 42.))
    assert ha.getNode('attr5').getType() == Types.COMPLEX_DOUBLE
    assert ha.get('attr5') == complex(12., 42.)
    # STRING
    ha.set('attr6', 'abrakadabra')
    assert ha.getNode('attr6').getType() == Types.STRING
    assert ha.get('attr6') == 'abrakadabra'

    # BOOL
    ha['attr7'] = True
    assert ha.getNode('attr7').getType() == Types.BOOL
    assert ha['attr7']
    # INT32
    ha['attr8'] = 42
    assert ha.getNode('attr8').getType(), Types.INT32
    assert ha['attr8'] == 42
    # INT64
    ha['attr9'] = -123456789012
    assert ha.getNode('attr9').getType() == Types.INT64
    assert ha['attr9'] == -123456789012
    # DOUBLE
    ha['attr10'] = 2.718281828
    assert ha.getNode('attr10').getType() == Types.DOUBLE
    assert ha['attr10'] == 2.718281828
    # COMPLEX_DOUBLE
    ha['attr11'] = complex(9999.12, -65432.03)
    assert ha.getNode('attr11').getType() == Types.COMPLEX_DOUBLE
    # STRING
    ha['attr12'] = 'some string'
    assert ha.getNode('attr12').getType() == Types.STRING
    # VECTOR_BOOL
    ha['attr13'] = [True, False, False, True, True]
    assert ha.getNode('attr13').getType() == Types.VECTOR_BOOL
    # VECTOR_INT32
    ha['attr14'] = [1, 2, 3, 4, 5]
    assert ha.getNode('attr14').getType() == Types.VECTOR_INT32
    # VECTOR_INT64
    ha['attr15'] = [12, 13, 1444444444444, -2, 0]
    assert ha.getNode('attr15').getType() == Types.VECTOR_INT64
    # VECTOR_DOUBLE
    ha['attr16'] = [1.11111, 2.222222, 3.333333]
    assert ha.getNode('attr16').getType() == Types.VECTOR_DOUBLE
    # VECTOR_COMPLEX_DOUBLE
    ha['attr17'] = [complex(1.0, -5), complex(-2., 188)]
    assert ha.getNode('attr17').getType() == Types.VECTOR_COMPLEX_DOUBLE
    assert ha.getType('attr17') == Types.VECTOR_COMPLEX_DOUBLE
    assert ha.isType('attr17', Types.VECTOR_COMPLEX_DOUBLE)
    assert ha.isType('attr17', "VECTOR_COMPLEX_DOUBLE")
    ha.erase('attr17')
    ha['attr17'] = [complex(1.0, -5), complex(-2., 188)]
    del ha['attr17']

    # VECTOR_STRING
    ha['attr18'] = ['aaaa', 'bbb', 'cccccccc', 'www']
    assert ha.getNode('attr18').getType() == Types.VECTOR_STRING
    assert len(ha) == 17
    assert 'attr18' in ha
    assert 'attr17' not in ha
    assert not ha.empty()
    # Get reference to node
    node6 = ha.getNode('attr6')
    assert isinstance(node6, Node)
    assert node6.getKey() == 'attr6'
    assert node6.getType() == Types.STRING
    assert node6.getValue() == 'abrakadabra'
    assert ha.getType('attr1') == Types.BOOL
    assert ha.isType('attr1', Types.BOOL)
    assert ha.isType('attr1', "BOOL")
    assert ha.getType('attr2') == Types.INT32
    assert ha.isType('attr2', Types.INT32)
    assert ha.isType('attr2', "INT32")
    assert ha.getType('attr3') == Types.INT64
    assert ha.isType('attr3', Types.INT64)
    assert ha.isType('attr3', "INT64")
    assert ha.getType('attr4') == Types.DOUBLE
    assert ha.isType('attr4', Types.DOUBLE)
    assert ha.isType('attr4', "DOUBLE")
    assert ha.getType('attr5') == Types.COMPLEX_DOUBLE
    assert ha.isType('attr5', Types.COMPLEX_DOUBLE)
    assert ha.isType('attr5', "COMPLEX_DOUBLE")
    assert ha.getType('attr6') == Types.STRING
    assert ha.isType('attr6', Types.STRING)
    assert ha.isType('attr6', "STRING")
    assert ha.getType('attr13') == Types.VECTOR_BOOL
    assert ha.isType('attr13', Types.VECTOR_BOOL)
    assert ha.isType('attr13', "VECTOR_BOOL")
    assert ha.getType('attr14') == Types.VECTOR_INT32
    assert ha.isType('attr14', Types.VECTOR_INT32)
    assert ha.isType('attr14', "VECTOR_INT32")
    assert ha.getType('attr15') == Types.VECTOR_INT64
    assert ha.isType('attr15', Types.VECTOR_INT64)
    assert ha.isType('attr15', "VECTOR_INT64")
    assert ha.getType('attr16') == Types.VECTOR_DOUBLE
    assert ha.isType('attr16', Types.VECTOR_DOUBLE)
    assert ha.isType('attr16', "VECTOR_DOUBLE")
    assert ha.getType('attr18') == Types.VECTOR_STRING
    assert ha.isType('attr18', Types.VECTOR_STRING)
    assert ha.isType('attr18', "VECTOR_STRING")
    assert not ha.isType('attr15', "INT64")

    ha.clear()
    assert (ha.empty())


def test_iteration():
    ha = Attributes()
    ha['a'] = 8100000000000000
    ha['b'] = True
    ha['c'] = 2.333333333
    ha['s'] = 'Hello'

    it = iter(ha)
    n = next(it)
    assert isinstance(n, Node)
    assert n.getKey() == 'a'
    assert n.getType() == Types.INT64
    assert n.getValue() == 8100000000000000
    assert n.getValueAs(Types.STRING) == '8100000000000000'

    n = next(it)
    assert isinstance(n, Node)
    assert n.getKey() == 'b'
    assert n.getType() == Types.BOOL
    assert n.getValue() is True

    n = next(it)
    assert n.getKey() == 'c'
    assert n.getType() == Types.DOUBLE

    n = next(it)
    assert n.getKey() == 's'
    assert n.getType() == Types.STRING
    assert n.getValue() == 'Hello'

    with pytest.raises(StopIteration):
        n = next(it)


def test_changetype():
    ha = Attributes()
    ha['a'] = 65

    def current_type_is(reftype):
        assert ha.getType('a') == reftype

    current_type_is(Types.INT32)
    # Get as INT64
    assert ha.getAs('a', "INT64"), 65
    # Get as INT16
    assert ha.getAs('a', "INT16"), 65
    # Get as INT8
    assert ha.getAs('a', "INT8"), 65
    # Get as DOUBLE
    assert ha.getAs('a', "DOUBLE"), 65.0
    # Attempt to get as BOOL: Cast Exception
    with pytest.raises(RuntimeError):
        assert (ha.getAs('a', "BOOL"))
    # Get as COMPLEX_DOUBLE
    assert ha.getAs('a', "COMPLEX_DOUBLE") == complex(65, 0)
    # Get as STRING
    assert ha.getAs('a', "STRING") == '65'

    # Change the type ...
    current_type_is(Types.INT32)
    assert ha['a'] == 65
    # INT64
    ha.getNode('a').setType("INT64")
    current_type_is(Types.INT64)
    assert ha['a'] == 65
    # INT16
    ha.getNode('a').setType("INT16")
    current_type_is(Types.INT16)
    assert ha['a'] == 65
    # INT8
    ha.getNode('a').setType("INT8")
    current_type_is(Types.INT8)
    assert ha['a'] == 65
    # DOUBLE
    ha.getNode('a').setType("DOUBLE")
    current_type_is(Types.DOUBLE)
    assert ha['a'] == 65.0
    # COMPLEX_DOUBLE
    ha.getNode('a').setType("COMPLEX_DOUBLE")
    current_type_is(Types.COMPLEX_DOUBLE)
    assert ha['a'] == complex(65, 0)  # (65+0j)
    # STRING
    ha.getNode('a').setType("STRING")
    current_type_is(Types.STRING)
    assert ha['a'] == '(65,0)'



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import pytest

from karabo.bound import (
    FLOAT_ELEMENT, INT32_ELEMENT, NODE_ELEMENT, Schema, Types)


class ShapePosition:
    @staticmethod
    def expectedParameters(expected):
        (
            FLOAT_ELEMENT(expected).key("shapeX")
            .displayedName("Shape X")
            .description("X coordinate of shape")
            .assignmentOptional().defaultValue(0.0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected).key("shapeY")
            .displayedName("Shape Y")
            .description("Y coordinate of shape")
            .assignmentOptional().defaultValue(0.0)
            .reconfigurable()
            .commit(),
        )


class Circle:
    @staticmethod
    def expectedParameters(expected):
        (
            NODE_ELEMENT(expected).key("center")
            .displayedName("Center")
            .description("Center of circle")
            .appendParametersOf(ShapePosition)
            .commit(),

            FLOAT_ELEMENT(expected).key("radius")
            .displayedName("Radius")
            .description("Radius of circle")
            .assignmentOptional().defaultValue(1.0)
            .reconfigurable()
            .commit(),
        )


class Rectangle:
    @staticmethod
    def expectedParameters(expected):
        (
            NODE_ELEMENT(expected).key("topLeftCorner")
            .displayedName("Top-left corner")
            .description("Top-left corner of rectangle")
            .appendParametersOf(ShapePosition)
            .commit(),

            FLOAT_ELEMENT(expected).key("width")
            .displayedName("Width")
            .description("Width of rectangle")
            .assignmentOptional().defaultValue(1.0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected).key("height")
            .displayedName("Height")
            .description("Height of rectangle")
            .assignmentOptional().defaultValue(1.0)
            .reconfigurable()
            .commit(),
        )


class A:
    @staticmethod
    def expectedParameters(expected):
        (
            INT32_ELEMENT(expected)
            .key("a")
            .assignmentOptional().defaultValue(10)
            .reconfigurable()
            .commit(),
        )


def test_appendSchema():
    dataschema = Schema()
    A.expectedParameters(dataschema)

    class B:
        @staticmethod
        def expectedParameters(expected):
            (
                NODE_ELEMENT(expected)
                .key("node")
                .appendSchema(dataschema)
                .commit(),
            )

    schema = Schema()
    B.expectedParameters(schema)
    assert schema.has("node.a") is True
    assert schema.getValueType("node.a") == Types.INT32
    assert schema.getDefaultValue("node.a") == 10


def test_appendParametersOf():
    schema = Schema("circle")
    Circle.expectedParameters(schema)
    assert schema.getRootName() == "circle"
    assert schema.isNode("center") is True
    assert schema.isLeaf("center.shapeX") is True
    assert schema.getValueType("center.shapeX") == Types.FLOAT
    assert schema.getDisplayedName("center.shapeX") == "Shape X"
    assert schema.getDisplayedName("center.shapeY") == "Shape Y"
    assert schema.getDefaultValue("center.shapeX") == 0.0
    assert schema.getDefaultValue("center.shapeY") == 0.0
    assert schema.getDisplayedName("radius") == "Radius"
    assert schema.getValueType("radius") == Types.FLOAT
    assert schema.getDefaultValue("radius") == 1.0
    schema = Schema("rect")
    Rectangle.expectedParameters(schema)
    assert schema.getRootName() == "rect"
    assert schema.isNode("topLeftCorner") is True
    assert schema.getDisplayedName("topLeftCorner.shapeX") == "Shape X"
    assert schema.getDisplayedName("topLeftCorner.shapeY") == "Shape Y"
    assert schema.getDefaultValue("topLeftCorner.shapeX") == 0.0
    assert schema.getDefaultValue("topLeftCorner.shapeY") == 0.0
    assert schema.getDisplayedName("width") == "Width"
    assert schema.getDisplayedName("height") == "Height"
    assert schema.getDefaultValue("width") == 1.0
    assert schema.getValueType("height") == Types.FLOAT
    assert schema.getDefaultValue("height") == 1.0


def test_keyValidation():
    s = Schema()
    NODE_ELEMENT(s).key("allowed").commit()
    NODE_ELEMENT(s).key("allowed", True).commit()
    with pytest.raises(RuntimeError, match="illegal character at position 3"):
        NODE_ELEMENT(s).key("Bad@Inside", True).commit()

    # We can tolerate the @, but not the $
    NODE_ELEMENT(s).key("Bad@Inside", False).commit()
    with pytest.raises(RuntimeError,
                       match="not tolerated character at position 3"):
        NODE_ELEMENT(s).key("Bad$Inside", False).commit()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import unittest

import numpy as np

from karabo.bound import (
    IMAGEDATA_ELEMENT, INT32_ELEMENT, NODE_ELEMENT, OVERWRITE_ELEMENT,
    AccessLevel, AccessType, ArchivePolicy, Assignment, Configurator,
    DaqDataType, Hash, ImageData, Logger, MetricPrefix, NodeType, Schema,
    State, Types, Unit, Validator, fullyEqual)

from .configuration_example_classes import (
    ArrayContainer, Base, GraphicsRenderer, GraphicsRenderer1, SomeClass,
    TestStruct1)


class Schema_TestCase(unittest.TestCase):

    def test_buildUp(self):
        try:
            # schema = Shape.getSchema("EditableCircle")
            schema = Configurator("Shape").getSchema("Circle")
            self.assertTrue(schema.isAccessInitOnly("shadowEnabled"))
            self.assertTrue(schema.isAccessInitOnly("radius"))
            self.assertTrue(schema.isLeaf("radius"))
            schema = Configurator("Shape").getSchema("EditableCircle")
            allowedStates = schema.getOptions("state")
            self.assertEqual(allowedStates, ['INIT', 'ERROR', 'NORMAL'])
            self.assertEqual(schema.getDefaultValue("state"), 'INIT')
            self.assertEqual(schema.getDefaultValue("stateN"), 'NORMAL')
            self.assertEqual(schema.getDefaultValue("stateE"), 'ERROR')
            self.assertEqual(schema.getDefaultValue("alarmW"), 'warn')
            self.assertEqual(schema.getDefaultValue("alarmA"), 'alarm')

            allowedStates = schema.getOptions("status")
            self.assertEqual(allowedStates, ['a', 'b', 'c'])
            self.assertEqual(schema.getDefaultValue("status"), 'a')
        except Exception as e:
            self.fail("test_buildUp exception group 1: " + str(e))

        try:
            schema = Schema("test")
            GraphicsRenderer1.expectedParameters(schema)
            self.assertTrue(schema.isAccessInitOnly("circle.radius"))
            self.assertTrue(schema.isLeaf("circle.radius"))
        except Exception as e:
            self.fail("test_buildUp exception group 2: " + str(e))

        try:
            instance = GraphicsRenderer.create("GraphicsRenderer",
                                               Hash("shapes", "Circle",
                                                    "Circle.radius", 0.5,
                                                    "color", "red",
                                                    "antiAlias", "true"))
            self.assertIsNotNone(instance)
        except Exception as e:
            self.fail("test_buildUp exception group 3: " + str(e))

    def test_getRootName(self):
        try:
            schema = Schema("MyTest")
            TestStruct1.expectedParameters(schema)
            self.assertEqual(schema.getRootName(), "MyTest")
        except Exception as e:
            self.fail("test_getRootName exception: " + str(e))

    def test_getTags(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getTags("exampleKey1")[0], "hardware")
            self.assertEqual(schema.getTags("exampleKey1")[1], "poll")
            self.assertEqual(schema.getTags("exampleKey2")[0], "hardware")
            self.assertEqual(schema.getTags("exampleKey2")[1], "poll")
            self.assertEqual(schema.getTags("exampleKey3")[0], "hardware")
            self.assertEqual(schema.getTags("exampleKey3")[1], "set")
            self.assertEqual(schema.getTags("exampleKey4")[0], "software")
            self.assertEqual(schema.getTags("exampleKey5")[0], "h/w")
            self.assertEqual(schema.getTags("exampleKey5")[1], "d.m.y")
        except Exception as e:
            self.fail("test_getTags exception: " + str(e))

    def test_setTags(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertTrue(schema.hasTags('x'))
            self.assertEqual(schema.getTags('x'), ['IK', 'BH'])
            schema.setTags('x', 'CY,SE')
            self.assertEqual(schema.getTags('x'), ['CY', 'SE'])
        except Exception as e:
            self.fail("test_setTags exception: " + str(e))

    def test_getsetExpertLevel(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getRequiredAccessLevel('x'),
                             AccessLevel.EXPERT)
            self.assertEqual(schema.getRequiredAccessLevel('y'),
                             AccessLevel.OPERATOR)
            self.assertEqual(schema.getRequiredAccessLevel('a'),
                             AccessLevel.OBSERVER)

            schema.setRequiredAccessLevel('x', AccessLevel.EXPERT)
            schema.setRequiredAccessLevel('y', AccessLevel.OPERATOR)
            self.assertEqual(schema.getRequiredAccessLevel('x'),
                             AccessLevel.EXPERT)
            self.assertEqual(schema.getRequiredAccessLevel('y'),
                             AccessLevel.OPERATOR)
        except Exception as e:
            self.fail("test_setTags exception group 1: " + str(e))

        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey1'),
                             AccessLevel.OPERATOR)
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey2'),
                             AccessLevel.OPERATOR)
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey3'),
                             AccessLevel.EXPERT)
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey4'),
                             AccessLevel.EXPERT)
            # default for readOnly
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey5'),
                             AccessLevel.OBSERVER)
            # default for reconfigurable
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey10'),
                             AccessLevel.OPERATOR)
            # observerAccess in reconfigurable
            self.assertEqual(schema.getRequiredAccessLevel('exampleKey11'),
                             AccessLevel.OBSERVER)
        except Exception as e:
            self.fail("test_setTags exception group 2: " + str(e))

    def test_getNodeType(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            nodeType = schema.getNodeType("exampleKey1")
            self.assertEqual(nodeType, NodeType.LEAF)
            self.assertEqual(schema.getNodeType("exampleKey5"), NodeType.LEAF)
        except Exception as e:
            self.fail("test_getNodeType exception: " + str(e))

    def test_getValueType(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getValueType("exampleKey1"), Types.STRING)
            self.assertEqual(schema.getValueType("exampleKey2"), Types.INT32)
            self.assertEqual(schema.getValueType("exampleKey3"), Types.UINT32)
            self.assertEqual(schema.getValueType("exampleKey4"), Types.DOUBLE)
            self.assertEqual(schema.getValueType("exampleKey5"), Types.INT64)
            self.assertEqual(schema.getValueType("exampleKey7"),
                             Types.VECTOR_INT32)
            self.assertEqual(schema.getValueType("exampleKey8"),
                             Types.VECTOR_DOUBLE)
            self.assertEqual(schema.getValueType("exampleKey9"),
                             Types.VECTOR_STRING)
        except Exception as e:
            self.fail("test_getValueType exception: " + str(e))

    def test_getAliasAsString(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getAliasAsString("exampleKey2"), "10")
            self.assertEqual(schema.getAliasAsString("exampleKey3"), "5.5")

            self.assertEqual(schema.getAliasAsString("exampleKey4"),
                             "exampleAlias4")
            self.assertEqual(schema.getAliasAsString("exampleKey5"),
                             "exampleAlias5")
            self.assertEqual(schema.getAliasAsString("exampleKey6"),
                             "1193046,43724")
            self.assertEqual(schema.getAliasAsString("testPath"), "5")
        except Exception as e:
            self.fail("test_getAliasAsString exception: " + str(e))

    def test_keyHasAlias(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertFalse(schema.keyHasAlias("exampleKey1"))
            self.assertTrue(schema.keyHasAlias("exampleKey2"))
            self.assertTrue(schema.keyHasAlias("exampleKey3"))
            self.assertTrue(schema.keyHasAlias("exampleKey4"))
            self.assertTrue(schema.keyHasAlias("exampleKey5"))
            self.assertTrue(schema.keyHasAlias("exampleKey6"))
            self.assertTrue(schema.keyHasAlias("testPath"))
        except Exception as e:
            self.fail("test_keyHasAlias exception: " + str(e))

    def test_aliasHasKey(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertTrue(schema.aliasHasKey(10))
            self.assertTrue(schema.aliasHasKey(5.5))
            self.assertTrue(schema.aliasHasKey("exampleAlias4"))
            self.assertTrue(schema.aliasHasKey("exampleAlias5"))
            self.assertTrue(schema.aliasHasKey([0x00123456, 0x0000aacc]))
            self.assertFalse(schema.aliasHasKey(7))
            self.assertTrue(schema.aliasHasKey(5))
        except Exception as e:
            self.fail("test_aliasHasKey exception: " + str(e))

    def test_getAliasFromKey(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getAliasFromKey("exampleKey2"), 10)
            self.assertEqual(schema.getAliasFromKey("exampleKey3"), 5.5)
            self.assertEqual(schema.getAliasFromKey("exampleKey4"),
                             "exampleAlias4")
            self.assertEqual(schema.getAliasFromKey("exampleKey5"),
                             "exampleAlias5")
            self.assertEqual(schema.getAliasFromKey("exampleKey6"),
                             [0x00123456, 0x0000aacc])
            self.assertEqual(schema.getAliasFromKey("testPath"), 5)
        except Exception as e:
            self.fail("test_getAliasFromKey exception: " + str(e))

    def test_setAlias(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getAliasFromKey("x"), 10)
            schema.setAlias('x', 'abc')
            self.assertEqual(schema.getAliasFromKey("x"), 'abc')
            schema.setAlias('x', 99)
            self.assertEqual(schema.getAliasFromKey("x"), 99)
        except Exception as e:
            self.fail("test_setAlias exception: " + str(e))

    def test_getKeyFromAlias(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getKeyFromAlias(10), "exampleKey2")
            self.assertEqual(schema.getKeyFromAlias(5.5), "exampleKey3")
            self.assertEqual(schema.getKeyFromAlias("exampleAlias4"),
                             "exampleKey4")
            self.assertEqual(schema.getKeyFromAlias("exampleAlias5"),
                             "exampleKey5")
            self.assertEqual(schema.getKeyFromAlias([0x00123456, 0x0000aacc]),
                             "exampleKey6")
            self.assertEqual(schema.getKeyFromAlias(5), "testPath")
        except Exception as e:
            self.fail("test_KeyFromAlias exception: " + str(e))

    def test_getAccessMode(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getAccessMode("exampleKey1"),
                             AccessType.WRITE)
            self.assertEqual(schema.getAccessMode("exampleKey2"),
                             AccessType.INIT)
            self.assertEqual(schema.getAccessMode("exampleKey3"),
                             AccessType.WRITE)
            self.assertEqual(schema.getAccessMode("exampleKey4"),
                             AccessType.INIT)
            self.assertEqual(schema.getAccessMode("exampleKey5"),
                             AccessType.READ)
            self.assertEqual(schema.getAccessMode("exampleKey8"),
                             AccessType.READ)
            self.assertEqual(schema.getAccessMode("exampleKey10"),
                             AccessType.WRITE)
            self.assertEqual(schema.getAccessMode("testPath"),
                             AccessType.WRITE)
            self.assertEqual(schema.getAccessMode("testPath2"),
                             AccessType.READ)
            self.assertEqual(schema.getAccessMode("testPath3"),
                             AccessType.INIT)
        except Exception as e:
            self.fail("test_getAccessMode exception: " + str(e))

    def test_getAssignment(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getAssignment("exampleKey1"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("exampleKey2"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("exampleKey3"),
                             Assignment.MANDATORY)
            self.assertEqual(schema.getAssignment("exampleKey4"),
                             Assignment.INTERNAL)
            self.assertEqual(schema.getAssignment("exampleKey5"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("exampleKey8"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("exampleKey10"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("testPath"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("testPath2"),
                             Assignment.OPTIONAL)
            self.assertEqual(schema.getAssignment("testPath3"),
                             Assignment.MANDATORY)
        except Exception as e:
            self.fail("test_getAssignment exception: " + str(e))

    def test_setAssignment(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertTrue(schema.hasAssignment('x'))
            self.assertTrue(schema.isAssignmentOptional('x'))
            self.assertFalse(schema.isAssignmentMandatory('x'))
            self.assertEqual(schema.getAssignment('x'),
                             Assignment.OPTIONAL)
            schema.setAssignment('x', Assignment.MANDATORY)
            self.assertFalse(schema.isAssignmentOptional('x'))
            self.assertTrue(schema.isAssignmentMandatory('x'))
            self.assertEqual(schema.getAssignment('x'), Assignment.MANDATORY)
        except Exception as e:
            self.fail("test_setAssignment exception: " + str(e))

    def test_getOptions(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            options = schema.getOptions("exampleKey1")
            self.assertEqual(options[0], "Radio")
            self.assertEqual(options[1], "Air Condition")
            self.assertEqual(options[2], "Navigation")

            self.assertEqual(schema.getOptions("exampleKey2")[0], 5)
            self.assertEqual(schema.getOptions("exampleKey2")[1], 25)
            self.assertEqual(schema.getOptions("exampleKey2")[2], 10)

            self.assertEqual(schema.getOptions("exampleKey4")[0], 1.11)
            self.assertEqual(schema.getOptions("exampleKey4")[1], -2.22)
            self.assertEqual(schema.getOptions("exampleKey4")[2], 5.55)

            self.assertEqual(schema.getOptions("testPath")[0], "file1")
            self.assertEqual(schema.getOptions("testPath")[1], "file2")
        except Exception as e:
            self.fail("test_getOptions exception: " + str(e))

    def test_setOptions(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")

            options = schema.getOptions("x")
            self.assertEqual(options[0], 5)
            self.assertEqual(options[1], 25)
            self.assertEqual(options[2], 10)
            self.assertEqual(schema.getOptions("x"), [5, 25, 10])

            schema.setOptions('x', '20, 5, 11, 13, 25')
            options = schema.getOptions("x")
            self.assertEqual(options, [20, 5, 11, 13, 25])
        except Exception as e:
            self.fail("test_setOptions exception: " + str(e))

    def test_displayType(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            display = schema.getDisplayType("myNode")
            self.assertEqual(display, "WidgetNode")
        except Exception as e:
            self.fail("test_displayType exception: " + str(e))

    def test_getDefaultValue(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertEqual(schema.getDefaultValue("exampleKey1"),
                             "Navigation")
            self.assertEqual(schema.getDefaultValue("exampleKey2"), 10)
            self.assertEqual(
                schema.getDefaultValueAs("exampleKey2", Types.STRING), "10")
            self.assertEqual(schema.getDefaultValue("exampleKey5"), 1442244)
            self.assertEqual(
                schema.getDefaultValueAs("exampleKey5", Types.STRING),
                "1442244")

            self.assertEqual(schema.getDefaultValue("exampleKey6"), 1.11)
            self.assertEqual(
                schema.getDefaultValueAs("exampleKey6", Types.DOUBLE), 1.11)

            self.assertEqual(schema.getDefaultValue("exampleKey7"), [1, 2, 3])
            self.assertEqual(schema.getDefaultValue("exampleKey8"),
                             [1.1, 2.2, 3.3])
            self.assertEqual(schema.getDefaultValue("exampleKey9"),
                             ["Hallo", "World"])

            # 'readOnly'-element (vector as well) that does not specify
            #  'initialValue' has 'defaultValue' equal to string "" :
            self.assertEqual(schema.getDefaultValue("testPath2"), "")
            self.assertEqual(schema.getDefaultValue("vectInt"), [])

            self.assertEqual(schema.getDefaultValue("exampleIntKey"), 20)

            self.assertEqual(schema.getDefaultValue("exampleKey5"), 1442244)

            self.assertEqual(schema.getDefaultValue("exampleKey7"), [1, 2, 3])

            # readOnly default specified by 'defaultValue'. not 'initialValue':
            self.assertEqual(schema.getDefaultValue("exampleKey5b"), 42)
            self.assertEqual(schema.getDefaultValue("exampleKey7b"),
                             [11, 22, 33])

        except Exception as e:
            self.fail("test_getDefaultValue exception: " + str(e))

    def test_setDefaultValue(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertTrue(schema.isAssignmentOptional('x'))
            self.assertTrue(schema.hasDefaultValue('x'))
            self.assertEqual(schema.getDefaultValue("x"), 5)
            schema.setDefaultValue("x", 10)
            self.assertEqual(schema.getDefaultValue("x"), 10)
        except Exception as e:
            self.fail("test_setDefaultValue exception: " + str(e))

    def test_getAllowedStates(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            allowedStates = schema.getAllowedStates("exampleKey3")
            self.assertEqual(allowedStates[0], State.STARTED)
            self.assertEqual(allowedStates[1], State.STOPPED)
            self.assertEqual(schema.getAllowedStates("exampleKey3")[2],
                             State.NORMAL)

            self.assertEqual(schema.getAllowedStates("exampleKey7")[0],
                             State.STARTED)
            self.assertEqual(schema.getAllowedStates("exampleKey7")[1],
                             State.NORMAL)
        except Exception as e:
            self.fail("test_getAllowedStates exception: " + str(e))

    def test_getUnit(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getUnit("exampleKey2"), Unit.METER)
            self.assertEqual(schema.getUnitName("exampleKey2"), "meter")
            self.assertEqual(schema.getUnitSymbol("exampleKey2"), "m")
        except Exception as e:
            self.fail("test_getUnit exception: " + str(e))

    def test_setUnit(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getUnit("x"), Unit.AMPERE)
            schema.setUnit('x', Unit.METER)
            self.assertEqual(schema.getUnit("x"), Unit.METER)
            self.assertEqual(schema.getUnit("x"), Unit.METER)
            self.assertEqual(schema.getUnitName("x"), "meter")
            self.assertEqual(schema.getUnitSymbol("x"), "m")
        except Exception as e:
            self.fail("test_setUnit exception: " + str(e))

    def test_getMetricPrefix(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getMetricPrefix("exampleKey2"),
                             MetricPrefix.MILLI)
            self.assertEqual(schema.getMetricPrefixName("exampleKey2"),
                             "milli")
            self.assertEqual(schema.getMetricPrefixSymbol("exampleKey2"), "m")
        except Exception as e:
            self.fail("test_getMetricPrefix exception: " + str(e))

    def test_setMetricPrefix(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getMetricPrefix("x"), MetricPrefix.MILLI)
            schema.setMetricPrefix("x", MetricPrefix.MICRO)
            self.assertEqual(schema.getMetricPrefix("x"), MetricPrefix.MICRO)
            self.assertEqual(schema.getMetricPrefixName("x"), "micro")
            self.assertEqual(schema.getMetricPrefixSymbol("x"), "u")
        except Exception as e:
            self.fail("test_setMetricPrefix exception: " + str(e))

    def test_getMinIncMaxInc(self):
        try:
            schema = TestStruct1.getSchema("TestStruct1")
            self.assertEqual(schema.getMinInc("exampleKey2"), 5)
            self.assertEqual(schema.getMinIncAs("exampleKey2", Types.STRING),
                             "5")
            self.assertEqual(schema.getMaxInc("exampleKey2"), 25)
            self.assertEqual(schema.getMaxIncAs("exampleKey2", Types.STRING),
                             "25")
        except Exception as e:
            self.fail("test_getMinIncMaxInc exception: " + str(e))

    def test_setMinIncMaxInc(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getMinInc("x"), 5)
            self.assertEqual(schema.getMinIncAs("x", Types.STRING), "5")
            self.assertEqual(schema.getMaxInc("x"), 25)
            self.assertEqual(schema.getMaxIncAs("x", Types.STRING), "25")
            schema.setMinInc('x', 3)
            schema.setMaxInc('x', 30)
            self.assertEqual(schema.getMinInc("x"), 3)
            self.assertEqual(schema.getMinIncAs("x", Types.STRING), "3")
            self.assertEqual(schema.getMaxInc("x"), 30)
            self.assertEqual(schema.getMaxIncAs("x", Types.STRING), "30")
        except Exception as e:
            self.fail("test_setMinIncMaxInc exception: " + str(e))

    def test_getMinExcMaxExc(self):
        schema = TestStruct1.getSchema("TestStruct1")
        try:
            self.assertEqual(schema.getMinExc("exampleKey3"), 10)
            self.assertEqual(schema.getMinExc("exampleKey4"), -2.22)
        except Exception as e:
            self.fail("test_getMinExcMaxExc exception in getMinExc: " + str(e))

        try:
            self.assertEqual(schema.getMaxExc("exampleKey3"), 20)
            self.assertEqual(schema.getMaxExc("exampleKey4"), 5.55)
        except Exception as e:
            self.fail("test_getMinExcMaxExc exception in getMaxExc: " + str(e))

        try:
            self.assertEqual(schema.getMinExcAs("exampleKey3", Types.STRING),
                             "10")
            self.assertEqual(schema.getMinExcAs("exampleKey4", Types.STRING),
                             "-2.22")
        except Exception as e:
            self.fail(
                "test_getMinExcMaxExc exception in getMinExcAs: " + str(e))

        try:
            self.assertEqual(schema.getMaxExcAs("exampleKey3", Types.STRING),
                             "20")
            self.assertEqual(schema.getMaxExcAs("exampleKey4", Types.STRING),
                             "5.55")
        except Exception as e:
            self.fail(
                "test_getMinExcMaxExc exception in getMaxExcAs: " + str(e))

    def test_setMinExcMaxExc(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getMinExc("y"), 0)
            self.assertEqual(schema.getMaxExc("y"), 29)
            schema.setMinExc("y", 2)
            schema.setMaxExc("y", 30)
            self.assertEqual(schema.getMinExc("y"), 2)
            self.assertEqual(schema.getMaxExc("y"), 30)
        except Exception as e:
            self.fail("test_setMinExcMaxExc exception in getMinExc: " + str(e))

    def test_vectorElement(self):
        schema = Configurator(TestStruct1).getSchema("TestStruct1")
        self.assertEqual(schema.isAccessReadOnly("exampleKey7"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey7"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey10"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey11"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey12"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey14"), True)
        self.assertEqual(schema.hasDefaultValue("exampleKey15"), True)

        self.assertEqual(schema.getDefaultValue("exampleKey10"), [10, 20, 30])

        self.assertEqual(schema.getDefaultValue("exampleKey12"),
                         [1.1, -2.2, 3.3])

        self.assertEqual(schema.getDefaultValue("exampleKey11"), [10, 20, 30])
        self.assertEqual(
            schema.getDefaultValueAs("exampleKey11", Types.STRING),
            "10,20,30")
        self.assertEqual(
            schema.getDefaultValueAs("exampleKey11", Types.VECTOR_INT32),
            [10, 20, 30])

        self.assertEqual(schema.getDefaultValue("exampleKey14"),
                         ["Hallo", "World", "Test"])
        self.assertEqual(
            schema.getDefaultValueAs("exampleKey14", Types.STRING),
            "Hallo,World,Test")

        self.assertEqual(schema.getDefaultValue("exampleKey15"),
                         ["word1", "word2", "test"])
        self.assertEqual(
            schema.getDefaultValueAs("exampleKey15", Types.STRING),
            "word1,word2,test")

        self.assertEqual(schema.getMinSize("exampleKey10"), 2)
        self.assertEqual(schema.getMaxSize("exampleKey10"), 7)

        schema = Configurator(SomeClass).getSchema("SomeClassId")
        validator = Validator()
        configuration = Hash('somelist', [])
        arr = np.ones(shape=(110, 210), dtype=np.int16)
        configuration["myImageElement"] = ImageData(arr)
        ok, error, validated = validator.validate(schema, configuration)
        assert ok, error
        somelist = validated['somelist']
        somelist.append(99)
        somelist.append(55)
        configuration['somelist'] = somelist
        ok, error, validated = validator.validate(schema, configuration)
        if not ok:
            raise RuntimeError(error)
        self.assertIsNotNone(validated)

    def test_ndarrayElement(self):
        schema = Configurator(ArrayContainer).getSchema("ArrayContainer")

        assert schema.getDefaultValue("exampleKey16.shape") == [2, 3]
        assert schema.getDefaultValue("exampleKey17.shape") == [2, 5, 0]
        assert schema.getDefaultValue("exampleKey18.shape") == [3, 2, 1]

        assert schema.getUnit("exampleKey16.data") == Unit.DEGREE_CELSIUS
        assert schema.getMetricPrefix(
            "exampleKey16.data") == MetricPrefix.CENTI

        assert schema.isAccessReadOnly("exampleKey16")
        assert schema.isAccessReadOnly("exampleKey17")
        assert schema.isAccessReadOnly("exampleKey18")

    def test_getDisplayType(self):
        schema = TestStruct1.getSchema("TestStruct1")
        self.assertEqual(schema.getDisplayType("exampleBitsKey1"), "bin")
        self.assertEqual(schema.getDisplayType("exampleBitsKey2"),
                         "bin|10:In Error, 21:Busy, 35:HV On, 55:Crate On")
        self.assertEqual(schema.getDisplayType("exampleBitsKey3"), "oct")
        self.assertEqual(schema.getDisplayType("exampleBitsKey4"), "hex")
        self.assertEqual(schema.getDisplayType("exampleKey5"),
                         "Int64DisplayType")
        self.assertEqual(schema.getDisplayType("exampleKey8"), "Curve")
        self.assertEqual(schema.getDisplayType("exampleKey9"),
                         "TestDisplayType")

    def test_setDisplayType(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertFalse(schema.hasDisplayType('y'))
            schema.setDisplayType('y', 'blabla')
            self.assertTrue(schema.hasDisplayType('y'))
            self.assertEqual(schema.getDisplayType("y"), "blabla")
        except Exception as e:
            self.fail("test_setDisplayType exception: " + str(e))

    def test_isCommand(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertTrue(schema.isCommand("slotTest"))
        except Exception as e:
            self.fail("test_isCommand exception: " + str(e))

    def test_isProperty(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertFalse(schema.isProperty("slotTest"))
            self.assertTrue(schema.isProperty("testPath2"))

        except Exception as e:
            self.fail("test_isProperty exception: " + str(e))

    def test_hasArchivePolicy(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertTrue(schema.hasArchivePolicy("exampleKey5"))
            self.assertFalse(schema.hasArchivePolicy("exampleKey6"))
            self.assertFalse(schema.hasArchivePolicy("exampleKey7"))
            self.assertTrue(schema.hasArchivePolicy("exampleKey8"))
        except Exception as e:
            self.fail("test_hasArchivePolicy exception: " + str(e))

    def test_getArchivePolicy(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertEqual(schema.getArchivePolicy("exampleKey5"),
                             ArchivePolicy.EVERY_EVENT)
            self.assertEqual(schema.getArchivePolicy("exampleKey8"),
                             ArchivePolicy.NO_ARCHIVING)
        except Exception as e:
            self.fail("test_getArchivePolicy exception: " + str(e))

    def test_setArchivePolicy(self):
        try:
            schema = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema.getArchivePolicy("a"),
                             ArchivePolicy.NO_ARCHIVING)
            schema.setArchivePolicy('a', ArchivePolicy.EVERY_EVENT)
            self.assertEqual(schema.getArchivePolicy("a"),
                             ArchivePolicy.EVERY_EVENT)
        except Exception as e:
            self.fail("test_setArchivePolicy exception: " + str(e))

    def test_perKeyFunctionality(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            keys = schema.getKeys()
            for key in keys:
                if key == "exampleKey1":
                    self.assertTrue(schema.hasAssignment(key))
                    self.assertTrue(schema.isAssignmentOptional(key))
                    self.assertTrue(schema.hasDefaultValue(key))
                    self.assertTrue(schema.hasAccessMode(key))
                    self.assertTrue(schema.isAccessReconfigurable(key))
                    self.assertTrue(schema.hasOptions(key))
                    self.assertTrue(schema.hasTags(key))
                    self.assertFalse(schema.hasUnit(key))
                    self.assertFalse(schema.hasMetricPrefix(key))

                if key == "exampleKey2":
                    self.assertTrue(schema.hasDefaultValue(key))
                    self.assertTrue(schema.hasAccessMode(key))
                    self.assertTrue(schema.isAccessInitOnly(key))
                    self.assertTrue(schema.hasOptions(key))
                    self.assertTrue(schema.hasTags(key))
                    self.assertFalse(schema.hasAllowedStates(key))
                    self.assertTrue(schema.hasUnit(key))
                    self.assertTrue(schema.hasMetricPrefix(key))
                    self.assertTrue(schema.hasMinInc(key))
                    self.assertTrue(schema.hasMaxInc(key))

                if key == "exampleKey3":
                    self.assertTrue(schema.hasAssignment(key))
                    self.assertTrue(schema.isAssignmentMandatory(key))
                    self.assertFalse(schema.hasDefaultValue(key))
                    self.assertFalse(schema.hasOptions(key))
                    self.assertTrue(schema.hasAllowedStates(key))
                    self.assertTrue(schema.hasMinExc(key))
                    self.assertTrue(schema.hasMaxExc(key))

                if key == "exampleKey4":
                    self.assertFalse(schema.hasDefaultValue(key))
                    self.assertTrue(schema.isAssignmentInternal(key))
                    self.assertTrue(schema.hasAccessMode(key))
                    self.assertTrue(schema.isAccessInitOnly(key))

                if key == "exampleKey5":
                    self.assertTrue(schema.hasDefaultValue(key))
                    self.assertTrue(schema.hasAssignment(key))
                    self.assertTrue(schema.isAssignmentOptional(key))
                    self.assertTrue(schema.hasAccessMode(key))
                    self.assertTrue(schema.isAccessReadOnly(key))

        except Exception as e:
            self.fail("test_perKeyFunctionality exception group 1: " + str(e))

    def test_merge(self):
        try:
            schema = Configurator(Base).getSchema('P1')
            self.assertTrue("a" in schema)
            self.assertFalse("x" in schema)
            self.assertFalse("y" in schema)
            self.assertFalse("z" in schema)

            schema2 = Configurator(Base).getSchema('P2')
            self.assertTrue("x" in schema2)
            self.assertTrue("y" in schema2)
            self.assertTrue("z" in schema2)

            schema += schema2

            self.assertTrue("a" in schema)
            self.assertTrue("x" in schema)
            self.assertTrue("y" in schema)
            self.assertTrue("z" in schema)

        except Exception as e:
            self.fail("test_merge exception: " + str(e))

    def test_logger(self):
        config = Hash("level", "DEBUG")
        Logger.configure(config)
        root = Logger.getLogger()
        a1 = Logger.getLogger("a1")
        a1_a2 = Logger.getLogger("a1.a2")
        root.debug("ERROR")

        Logger.useConsole()
        root.debug("ok")
        a1.debug("ok")
        a1_a2.debug("ok")
        root.INFO("ok")
        a1.INFO("ok")
        a1_a2.INFO("ok")

        Logger.reset()
        root.debug("ERROR")
        a1.debug("ERROR")
        a1_a2.debug("ERROR")

    def test_helpFunction(self):
        pass
        # uncomment to see help:
        # schema = TestStruct1.getSchema("TestStruct1")
        # schema.help()

    def test_schemaImageElement(self):
        try:
            schema = Configurator(TestStruct1).getSchema("TestStruct1")
            self.assertEqual(schema.getDisplayType("myImageElement"),
                             "ImageData")
            self.assertEqual(schema.isCustomNode("myImageElement"), True)
            self.assertEqual(schema.getCustomNodeClass("myImageElement"),
                             "ImageData")
            self.assertEqual(schema.getAccessMode("myImageElement"),
                             AccessType.READ)
            self.assertEqual(schema.getNodeType("myImageElement"),
                             NodeType.NODE)
            self.assertEqual(schema.getRequiredAccessLevel("myImageElement"),
                             AccessLevel.OPERATOR)  # .operatorAccess()
            self.assertEqual(schema.getDisplayedName("myImageElement"),
                             "myImage")
            self.assertEqual(schema.getDescription("myImageElement"),
                             "Image Element")
            self.assertEqual(schema.getDefaultValue("myImageElement.dims"),
                             [100, 200])
            schema2 = Configurator(SomeClass).getSchema("SomeClassId")
            self.assertEqual(schema2.getDefaultValue("myImageElement.dims"),
                             [110, 210])
            ide = IMAGEDATA_ELEMENT(schema2).key("myImageElement2")
            with self.assertRaises(RuntimeError):
                ide.setDimensions(123)  # only list ans str are allowed

            # Hijack this test to test correct false result of isCustomNode,
            # but here just one type, excessive test is done for
            # underlying C++ code
            self.assertEqual(schema.isCustomNode("exampleKey1"), False)
        except Exception as e:
            self.fail("test_schemaImageElement group 1: " + str(e))

        try:
            # self.assertEqual(schema.getDescription("myImageElement.pixels"),
            #  "Pixel array")
            # self.assertEqual(schema.getValueType("myImageElement.pixels"),
            # Types.VECTOR_CHAR)

            self.assertEqual(schema.getDisplayedName("myImageElement.dims"),
                             "Dimensions")
            self.assertEqual(schema.getValueType("myImageElement.dims"),
                             Types.VECTOR_UINT64)
            self.assertEqual(schema.getDisplayType("myImageElement.dims"),
                             "Curve")

            self.assertEqual(
                schema.getDisplayedName("myImageElement.encoding"), "Encoding")
            self.assertEqual(schema.getValueType("myImageElement.encoding"),
                             Types.INT32)

        # self.assertEqual(
        # schema.getDisplayedName("myImageElement.pixels.isBigEndian"),
        # "Is big endian")
        # self.assertEqual(
        # schema.getValueType("myImageElement.pixels.isBigEndian"), Types.BOOL)
        # self.assertEqual(
        # schema.getDefaultValue("myImageElement.pixels.isBigEndian"), False)

        except Exception as e:
            self.fail("test_schemaImageElement group 2: " + str(e))

    def test_allowed_states(self):
        schema = Configurator(TestStruct1).getSchema("TestStruct1")
        schema.setAllowedStates("exampleKey3", (State.INIT, State.NORMAL))
        allowedStates = schema.getAllowedStates("exampleKey3")
        self.assertEqual(allowedStates, [State.INIT, State.NORMAL])

    def test_daq_data_type(self):
        schema = Schema()

        NODE_ELEMENT(schema).key(
            "trainData").commit()  # Has no DAQ data type yet
        NODE_ELEMENT(schema).key("pulseData").setDaqDataType(
            DaqDataType.PULSE).commit()

        self.assertEqual(schema.hasDaqDataType("trainData"), False)
        self.assertEqual(schema.hasDaqDataType("pulseData"), True)
        self.assertEqual(schema.getDaqDataType("pulseData"), DaqDataType.PULSE)

        # Now add DAQ data type to node "trainData"
        schema.setDaqDataType("trainData", DaqDataType.TRAIN)
        self.assertEqual(schema.hasDaqDataType("trainData"), True)
        self.assertEqual(schema.getDaqDataType("trainData"), DaqDataType.TRAIN)

    def test_table_element(self):
        schema = TestStruct1.getSchema("TestStruct1")

        # Both, defaultValue and initialValue should work for read-only tables
        default = schema.getDefaultValue("tableI")
        self.assertEqual(len(default), 1)
        self.assertTrue(fullyEqual(default[0], Hash("int", 2)))
        default = schema.getDefaultValue("tableD")
        self.assertEqual(len(default), 1)
        self.assertTrue(fullyEqual(default[0], Hash("int", 3)))

        # There are more tests of tables in the integration tests...

    def test_overwrite_restrictions_for_options(self):
        schema = Schema()
        (
            INT32_ELEMENT(schema).key("range")
            .displayedName("Range")
            .options("0,1")
            .assignmentOptional().defaultValue(0)
            .reconfigurable()
            .commit()
        )
        vec = schema.getOptions("range")
        self.assertEqual(len(vec), 2)
        self.assertEqual(vec[0], 0)
        self.assertEqual(vec[1], 1)
        (
            OVERWRITE_ELEMENT(schema).key("range")
            .setNewOptions("0,1,2")
            .commit()
        )
        vec = schema.getOptions("range")
        self.assertEqual(len(vec), 3)
        self.assertEqual(vec[0], 0)
        self.assertEqual(vec[1], 1)
        self.assertEqual(vec[2], 2)

    def test_overwrite_tags(self):
        schema = Schema()
        (
            INT32_ELEMENT(schema).key("taggedProp")
            .tags("bip, bop")
            .assignmentOptional().defaultValue(0)
            .reconfigurable()
            .commit()
        )
        vec = schema.getTags("taggedProp")
        self.assertEqual(len(vec), 2)
        self.assertEqual(vec[0], "bip")
        self.assertEqual(vec[1], "bop")
        (
            OVERWRITE_ELEMENT(schema).key("taggedProp")
            .setNewTags("doff")
            .commit()
        )
        vec = schema.getTags("taggedProp")
        self.assertEqual(len(vec), 1)
        self.assertEqual(vec[0], "doff")
        (
            OVERWRITE_ELEMENT(schema).key("taggedProp")
            .setNewTags(["doff", "deff"])
            .commit()
        )
        vec = schema.getTags("taggedProp")
        self.assertEqual(len(vec), 2)
        self.assertEqual(vec[0], "doff")
        self.assertEqual(vec[1], "deff")
        (
            OVERWRITE_ELEMENT(schema).key("taggedProp")
            .setNewTags(("chip", "chop"))
            .commit()
        )
        vec = schema.getTags("taggedProp")
        self.assertEqual(len(vec), 2)
        self.assertEqual(vec[0], "chip")
        self.assertEqual(vec[1], "chop")
        # Only iterable of str are allowed
        with self.assertRaises(RuntimeError):
            (
                OVERWRITE_ELEMENT(schema).key("taggedProp")
                .setNewTags((1, 2))
                .commit()
            )


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import Hash, HashAttributes, HashAttributesNode, Types


def test_hashAttributesNode_getValueAs():
    # Test "getKey", "getValue", "getType", "getValueAs",
    #      "setType", "setValue"
    h = Hash("a.b.c", "value")
    assert isinstance(h, Hash)
    h.setAttribute("a.b.c", "attr1", 10)
    h.setAttribute("a.b.c", "attr2", [True, False])
    attrs = h.getAttributes("a.b.c")
    assert isinstance(attrs, HashAttributes)
    n1 = attrs.getNode("attr1")
    assert isinstance(n1, HashAttributesNode)
    n2 = attrs.getNode("attr2")

    # Test "getKey"
    assert n1.getKey() == "attr1"
    assert n2.getKey() == "attr2"

    # Test "getType"
    assert n1.getType() == Types.INT32
    assert n2.getType() == Types.VECTOR_BOOL

    # Test "getValue"
    assert n1.getValue() == 10
    assert n2.getValue() == [True, False]

    # Test "getValueAs"
    assert n1.getValueAs(Types.INT8) == 10
    assert n1.getValueAs(Types.UINT8) == 10
    assert n1.getValueAs(Types.INT16) == 10
    assert n1.getValueAs(Types.UINT16) == 10
    assert n1.getValueAs(Types.INT32) == 10
    assert n1.getValueAs(Types.UINT32) == 10
    assert n1.getValueAs(Types.INT64) == 10
    assert n1.getValueAs(Types.UINT64) == 10
    assert n1.getValueAs(Types.FLOAT) == 10
    assert n1.getValueAs(Types.DOUBLE) == 10
    assert n1.getValueAs(Types.COMPLEX_FLOAT) == (10+0j)
    assert n1.getValueAs(Types.COMPLEX_DOUBLE) == (10+0j)
    assert n1.getValueAs(Types.STRING) == '10'
    assert n1.getValueAs("STRING") == '10'
    # Raises an exception:
    # assert n2.getValueAs("VECTOR_CHAR") == bytearray(b'\x01\x00')
    assert n2.getValueAs("VECTOR_INT8") == [1, 0]
    assert n2.getValueAs("VECTOR_UINT8") == [1, 0]
    assert n2.getValueAs("VECTOR_INT16") == [1, 0]
    assert n2.getValueAs("VECTOR_UINT16") == [1, 0]
    assert n2.getValueAs("VECTOR_INT32") == [1, 0]
    assert n2.getValueAs("VECTOR_UINT32") == [1, 0]
    assert n2.getValueAs("VECTOR_INT64") == [1, 0]
    assert n2.getValueAs("VECTOR_UINT64") == [1, 0]
    assert n2.getValueAs("VECTOR_FLOAT") == [1.0, 0.0]
    assert n2.getValueAs("VECTOR_DOUBLE") == [1.0, 0.0]
    assert n2.getValueAs("VECTOR_COMPLEX_FLOAT") == [(1+0j), (0+0j)]
    assert n2.getValueAs("VECTOR_COMPLEX_DOUBLE") == [(1+0j), (0+0j)]
    assert n2.getValueAs("VECTOR_STRING") == ['1', '0']

    # Test 'setType' & 'setValue' for INT8
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("INT8")
    assert n1.getType() == Types.INT8
    assert n1.getValue() == 10
    n1.setValue(99)
    assert n1.getValue() == 99

    # 'setType' & 'setValue' for UINT8
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("UINT8")
    assert n1.getType() == Types.UINT8
    assert n1.getValue() == 10
    n1.setValue(13)
    assert n1.getValue() == 13

    # 'setType' & 'setValue' for INT16
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("INT16")
    assert n1.getType() == Types.INT16
    assert n1.getValue() == 10
    n1.setValue(3000)
    assert n1.getValue() == 3000

    # 'setType' & 'setValue' for UINT16
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("UINT16")
    assert n1.getType() == Types.UINT16
    assert n1.getValue() == 10
    n1.setValue(64000)
    assert n1.getValue() == 64000

    # 'setType' & 'setValue' for INT32
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("INT32")
    assert n1.getType() == Types.INT32
    assert n1.getValue() == 10
    n1.setValue(1_000_000_000)
    assert n1.getValue() == 1_000_000_000

    # 'setType' & 'setValue' for UINT32
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("UINT32")
    assert n1.getType() == Types.UINT32
    assert n1.getValue() == 10
    n1.setValue(3_000_000_000)
    assert n1.getValue() == 3_000_000_000

    # 'setType' & 'setValue' for INT64
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("INT64")
    assert n1.getType() == Types.INT64
    assert n1.getValue() == 10
    n1.setValue(9_223_372_036_854_775_807)
    assert n1.getValue() == 9_223_372_036_854_775_807

    # 'setType' & 'setValue' for UINT64
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("UINT64")
    assert n1.getType() == Types.UINT64
    assert n1.getValue() == 10
    n1.setValue(18_446_744_073_709_551_600)
    assert n1.getValue() == 18_446_744_073_709_551_600

    # 'setType' & 'setValue' for FLOAT
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("FLOAT")
    assert n1.getType() == Types.FLOAT
    assert n1.getValue() == 10
    n1.setValue(-3.141592)
    assert n1.getValue() == -3.141592

    # 'setType' & 'setValue' for DOUBLE
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("DOUBLE")
    assert n1.getType() == Types.DOUBLE
    assert n1.getValue() == 10
    n1.setValue(2345.123456799)
    assert n1.getValue() == 2345.123456799

    # 'setType' & 'setValue' for COMPLEX_FLOAT
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("COMPLEX_FLOAT")
    assert n1.getType() == Types.COMPLEX_FLOAT
    assert n1.getValue() == (10+0j)
    n1.setValue(12+9.9j)
    assert n1.getValue() == (12+9.9j)

    # 'setType' & 'setValue' for COMPLEX_DOUBLE
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("COMPLEX_DOUBLE")
    assert n1.getType() == Types.COMPLEX_DOUBLE
    assert n1.getValue() == (10+0j)
    n1.setValue(2.222222222-33.3333333333j)
    assert n1.getValue() == (2.222222222-33.3333333333j)

    attrs = h.getAttributes("a.b.c")
    # 'setType' & 'setValue' for STRING
    h.setAttribute("a.b.c", "attr1", 10)
    n1 = attrs.getNode("attr1")
    n1.setType(Types.STRING)
    assert n1.getType() == Types.STRING
    assert n1.getValue() == '10'
    n1.setValue('3_000_000_000')
    assert n1.getValue() == '3_000_000_000'

    # 'setType' VECTOR_INT8
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_INT8")
    assert n1.getType() == Types.VECTOR_INT8
    assert n1.getValue() == [10]

    # 'setType' VECTOR_UINT8
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_UINT8")
    assert n1.getType() == Types.VECTOR_UINT8
    assert n1.getValue() == [10]

    # 'setType' VECTOR_INT16
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_INT16")
    assert n1.getType() == Types.VECTOR_INT16
    assert n1.getValue() == [10]
    n1.setValue([8192, -16000])
    assert n1.getValue() == [8192, -16000]

    # 'setType' VECTOR_UINT16
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_UINT16")
    assert n1.getType() == Types.VECTOR_UINT16
    assert n1.getValue() == [10]
    n1.setValue([64000, 61000])
    assert n1.getValue() == [64000, 61000]

    # 'setType' VECTOR_INT32
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_INT32")
    assert n1.getType() == Types.VECTOR_INT32
    assert n1.getValue() == [10]
    n1.setValue([20064000, 1333332, 12, -402345])
    assert n1.getValue() == [20064000, 1333332, 12, -402345]

    # 'setType' VECTOR_UINT32
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_UINT32")
    assert n1.getType() == Types.VECTOR_UINT32
    assert n1.getValue() == [10]
    n1.setValue([1, 2, 3, 4, 5])
    assert n1.getValue() == [1, 2, 3, 4, 5]

    # 'setType' VECTOR_INT64
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_INT64")
    assert n1.getType() == Types.VECTOR_INT64
    assert n1.getValue() == [10]
    n1.setValue([-10064000, 2999999999999])
    assert n1.getValue() == [-10064000, 2999999999999]

    # 'setType' VECTOR_UINT64
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_UINT64")
    assert n1.getType() == Types.VECTOR_UINT64
    assert n1.getValue() == [10]
    n1.setValue([478564000])
    assert n1.getValue() == [478564000]

    # 'setType' VECTOR_FLOAT
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_FLOAT")
    assert n1.getType() == Types.VECTOR_FLOAT
    assert n1.getValue() == [10.0]
    n1.setValue([3.14])
    assert n1.getValue() == [3.14]

    # 'setType' VECTOR_DOUBLE
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_DOUBLE")
    assert n1.getType() == Types.VECTOR_DOUBLE
    assert n1.getValue() == [10.0]
    n1.setValue([111.22222, 33333.4444])
    assert n1.getValue() == [111.22222, 33333.4444]

    # 'setType' VECTOR_COMPLEX_FLOAT
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_COMPLEX_FLOAT")
    assert n1.getType() == Types.VECTOR_COMPLEX_FLOAT
    assert n1.getValue() == [(10+0j)]
    n1.setValue([(1.2-3.2j), (5.4+3.3j)])
    assert n1.getValue() == [(1.2-3.2j), (5.4+3.3j)]

    # 'setType' VECTOR_COMPLEX_DOUBLE
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType("VECTOR_COMPLEX_DOUBLE")
    assert n1.getType() == Types.VECTOR_COMPLEX_DOUBLE
    assert n1.getValue() == [(10+0j)]
    n1.setValue([(1.2-3.2j), (5.4+3.3j)])
    assert n1.getValue() == [(1.2-3.2j), (5.4+3.3j)]

    # 'setType' VECTOR_STRING
    h.setAttribute("a.b.c", "attr1", 10)
    attrs = h.getAttributes("a.b.c")
    n1 = attrs.getNode("attr1")
    n1.setType(Types.VECTOR_STRING)
    assert n1.getType() == Types.VECTOR_STRING
    assert n1.getValue() == ['10']
    n1.setValue(['cast', 'to', 'python', 'type'])
    assert n1.getValue() == ['cast', 'to', 'python', 'type']

    h.setAttribute("a.b.c", "attr2", "test")
    n2 = attrs.getNode("attr2")
    assert n2.getType() == Types.STRING
    n1.setValue(['karabim', 'karabom'])
    assert n1.getValue() == ['karabim', 'karabom']



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import unittest

from karabo.bound import Hash, PluginLoader


class PluginLoader_TestCase(unittest.TestCase):

    def test_default_namespace(self):
        loader = PluginLoader.create("PythonPluginLoader", Hash())
        entrypoints = loader.update()
        assert len(entrypoints) > 0, str(entrypoints)

    def test_broken_namespace(self):
        loaderCfg = Hash("pluginNamespace", "karabo.bound_broken_device_test")
        loader = PluginLoader.create("PythonPluginLoader", loaderCfg)
        entrypoints = loader.update()
        assert len(entrypoints) == 0
        assert len(loader.plugin_errors) > 0



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# To change this template, choose Tools | Templates
# and open the template in the editor.

import platform
import time
import unittest

from karabo.bound import (
    ATTOSEC, FEMTOSEC, MICROSEC, MILLISEC, NANOSEC, PICOSEC, Epochstamp, Hash,
    TimeDuration, TimeId, Timestamp, Types)


class Timestamp_TestCase(unittest.TestCase):
    # def setUp(self):
    #    self.foo = Timestamp_()
    #

    # def tearDown(self):
    #    self.foo.dispose()
    #    self.foo = None

    def test_timestamp_constructors(self):

        functionName = "test_timestamp_constructors"
        sleepDelay = 1

        try:
            ts01 = Timestamp()
        except Exception as e:
            self.fail(
                functionName + " creating ts01 default constructor: " + str(e))

        time.sleep(sleepDelay)

        try:
            ts02 = Timestamp()
        except Exception as e:
            self.fail(
                functionName + " creating ts02 default constructor: " + str(e))

        # The diff will be a tiny bit larger than sleepDelay, and depending
        # on the fractional seconds of ts01, the full seconds may differ
        # by one more:
        self.assertTrue(
            ts02.getSeconds() - ts01.getSeconds() <= sleepDelay + 1)

        self.assertNotEqual(ts01.getFractionalSeconds(),
                            ts02.getFractionalSeconds(),
                            "1st and 2nd timestamp should have differents"
                            " fractional seconds")

    def test_timestamp_to_iso8601_string(self):

        functionName = "test_timestamp_to_iso8601_string"
        # ISO8601 compact version
        pTimeStr01 = "20121225T132536.789333123456789123"
        self.assertIsNotNone(pTimeStr01)
        localeNameUS = "en_US.UTF-8"

        try:
            es01 = Epochstamp(1356441936, 789333123456789123)  # (pTimeStr01)
        except Exception as e:
            self.fail(
                f"{functionName} creating Epochstamp01 using the String"
                f" constructor: {str(e)}")

        try:
            ts01 = Timestamp(es01, TimeId())

            #####
            # Function toFormattedStringLocale and toFormattedString only
            # differ in the used locale, since internally
            # both use the same function: toFormattedStringInternal
            #
            # toFormattedStringLocale - requires that the locale name
            # toFormattedString - uses the System locale
            #
            # In this test only the toFormattedStringLocale, since this way
            # it's possible to run successfully this test
            # in Systems with a different locale
            #####

            # This test can only be run in LINUX since C++ locale support
            # seems completely broken on MAC OSX
            # The solution to make locale works in OSX passes for CLANG use
            if platform.system() == 'Darwin':
                ptime_osx_simple_str = '2012-Dec-25 13:25:36.789333'

                # Validate that "UNIVERSAL" format is corrected generated
                pTimeConvertedStr01 = ts01.toFormattedStringLocale(
                    localeNameUS, "%Y%m%dT%H%M%S")
                self.assertEqual(pTimeConvertedStr01, ptime_osx_simple_str,
                                 "These strings must be equal")

                # Validate that default "user-friendly" format is
                # corrected generated
                pTimeConvertedStr02 = ts01.toFormattedStringLocale(
                    localeNameUS)
                self.assertEqual(pTimeConvertedStr02, ptime_osx_simple_str,
                                 "These strings must be equal")

                # Validate that INPUT "user-friendly" format is
                # corrected generated
                pTimeConvertedStr03 = ts01.toFormattedStringLocale(
                    localeNameUS, "%Y/%m/%d %H:%M:%S")
                self.assertEqual(pTimeConvertedStr03, ptime_osx_simple_str,
                                 "These strings must be equal")

            else:
                # Validate that "UNIVERSAL" format is corrected generated
                pTimeConvertedStr01 = ts01.toFormattedStringLocale(
                    localeNameUS, "%Y%m%dT%H%M%S")
                self.assertEqual(pTimeConvertedStr01,
                                 "20121225T132536.789333",
                                 "These strings must be equal")

                # Validate that default "user-friendly" format is
                # corrected generated
                pTimeConvertedStr02 = ts01.toFormattedStringLocale(
                    localeNameUS)
                self.assertEqual(pTimeConvertedStr02,
                                 "2012-Dec-25 13:25:36.789333",
                                 "These strings must be equal")

                # Validate that INPUT "user-friendly" format is
                # corrected generated
                pTimeConvertedStr03 = ts01.toFormattedStringLocale(
                    localeNameUS, "%Y/%m/%d %H:%M:%S")
                self.assertEqual(pTimeConvertedStr03,
                                 "2012/12/25 13:25:36.789333",
                                 "These strings must be equal")

        except Exception as e:
            self.fail(
                " creating Timestamp01 using the String constructor: "
                + str(e))

        # Validate time was correctly converted/stored with the
        # correct information
        self.assertEqual(ts01.getSeconds(), 1356441936,
                         "1st timestamp number of seconds must be 1356441936")
        self.assertEqual(ts01.getFractionalSeconds(), 789333123456789123,
                         "1st timestamp number of fractional seconds must be"
                         " 789333123456789123")

    def test_timeduration(self):

        try:
            durZero = TimeDuration()
            self.assertEqual(durZero.getSeconds(), 0)
            self.assertEqual(durZero.getFractions(), 0)
        except Exception as e:
            self.fail(
                " testing default constructed time duration to be zero: "
                + str(e))

        try:
            seconds = 3600  # one hour
            fractionsAtto = 4565460000000  # 456.546 micro seconds

            dur1 = TimeDuration(seconds, fractionsAtto)
            self.assertEqual(dur1.getSeconds(), 0)
            self.assertEqual(dur1.getTotalSeconds(), seconds)
            self.assertEqual(dur1.getMinutes(), 0)
            self.assertEqual(dur1.getTotalMinutes(), 60)
            self.assertEqual(dur1.getHours(), 1)
            self.assertEqual(dur1.getTotalHours(), 1)
            self.assertEqual(dur1.getFractions(ATTOSEC), fractionsAtto)
            self.assertEqual(dur1.getFractions(FEMTOSEC),
                             fractionsAtto // 1000)
            self.assertEqual(dur1.getFractions(PICOSEC),
                             fractionsAtto // 1000000)
            self.assertEqual(dur1.getFractions(NANOSEC),
                             fractionsAtto // 1000000000)
            self.assertEqual(dur1.getFractions(MICROSEC),
                             fractionsAtto // 1000000000000)
            self.assertEqual(dur1.getFractions(MILLISEC),
                             fractionsAtto // 1000000000000000)

        except Exception as e:
            self.fail(
                " testing construction from seconds and fractions: " + str(e))

        try:
            h = Hash()
            # Cannot use Hash("seconds", 987, "fractions", 123),
            # but have to set explicitely with type, otherwise Hash
            # uses int32...
            h.setAs("seconds", 987, Types.UINT64)
            h.setAs("fractions", 123, Types.UINT64)
            dur2 = TimeDuration(h)
            dur3 = TimeDuration(987, 123)
            self.assertEqual(dur3 - dur2, durZero)
        except Exception as e:
            self.fail(" testing construction from hash: " + str(e))

        try:
            # Test equal comparisons
            durA = TimeDuration(123, 4567890000)
            durB = TimeDuration(123, 4567890000)
            self.assertTrue(durA == durB)
            self.assertTrue(durA <= durB)
            self.assertTrue(durA >= durB)
            self.assertFalse(durA != durB)

            # Test larger/smaller comparisons with equal seconds
            durC = TimeDuration(123, 4567890000)
            durD = TimeDuration(123, 4567890001)
            self.assertTrue(durC != durD)
            self.assertTrue(durC < durD)
            self.assertTrue(durC <= durD)
            self.assertTrue(durD > durC)
            self.assertTrue(durD >= durC)
            self.assertFalse(durD < durC)
            self.assertFalse(durD <= durC)
            self.assertFalse(durC > durD)
            self.assertFalse(durC >= durD)

            # Test larger/smaller comparisons with equal fractions
            durE = TimeDuration(3, 4567890000)
            durF = TimeDuration(4, 4567890000)
            self.assertTrue(durE != durF)
            self.assertTrue(durE < durF)
            self.assertTrue(durE <= durF)
            self.assertTrue(durF > durE)
            self.assertTrue(durF >= durE)
            self.assertFalse(durF < durE)
            self.assertFalse(durF <= durE)
            self.assertFalse(durE > durF)
            self.assertFalse(durE >= durF)

            # Test larger/smaller comparisons with seconds smaller,
            # fractions larger
            durG = TimeDuration(444, 4567890000)
            durH = TimeDuration(555, 1234560000)
            self.assertTrue(durG != durH)
            self.assertTrue(durG < durH)
            self.assertTrue(durG <= durH)
            self.assertTrue(durH > durG)
            self.assertTrue(durH >= durG)
            self.assertFalse(durH < durG)
            self.assertFalse(durH <= durG)
            self.assertFalse(durG > durH)
            self.assertFalse(durG >= durH)

            # Skip here testing operator+, operator- and operator/,
            # see C++ tests.

        except Exception as e:
            self.fail(" testing comparison operators: " + str(e))

    def test_time_attributes(self):

        try:
            # Test time attributes
            h = Hash()
            h["timestamp"] = True
            h.setAttribute("timestamp", "tid", 22)
            h.setAttribute("timestamp", "sec", 1234)
            h.setAttribute("timestamp", "frac", 5678)
            attrs = h.getAttributes("timestamp")
            self.assertEqual(
                Timestamp.hashAttributesContainTimeInformation(attrs), True)

            tm = Timestamp.fromHashAttributes(attrs)

            self.assertEqual(tm.getTid(), 22)
            self.assertEqual(tm.getSeconds(), 1234)
            self.assertEqual(tm.getFractionalSeconds(), 5678)
        except Exception as e:
            self.fail(" testing conversion from attributes: " + str(e))

        try:
            # Test Train ID upgrade to uint64 without explicit Types.UINT64
            h = Hash()
            h["timestamp"] = True
            h.setAttribute("timestamp", "tid", 2 ** 40)
            h.setAttribute("timestamp", "sec", 1234)
            h.setAttribute("timestamp", "frac", 5678)

            attrs = h.getAttributes("timestamp")
            self.assertEqual(
                Timestamp.hashAttributesContainTimeInformation(attrs), True)
            tm = Timestamp.fromHashAttributes(attrs)

            self.assertEqual(tm.getTid(), 2 ** 40)
            self.assertEqual(tm.getSeconds(), 1234)
            self.assertEqual(tm.getFractionalSeconds(), 5678)
        except Exception as e:
            self.fail(
                " testing conversion from attributes with large train id: "
                + str(e))

        try:
            # Test explicit time attributes set as uint64
            h64 = Hash()
            h64.setAs("tid", 2 ** 40, Types.UINT64)
            h64.setAs("sec", 2 ** 41, Types.UINT64)
            h64.setAs("frac", 2 ** 42, Types.UINT64)

            h["timestamp64"] = True
            h.setAttribute("timestamp64", "tid", h64["tid"])
            h.setAttribute("timestamp64", "sec", h64["sec"])
            h.setAttribute("timestamp64", "frac", h64["frac"])
            attrs = h.getAttributes("timestamp64")
            self.assertEqual(
                Timestamp.hashAttributesContainTimeInformation(attrs), True)
            tm = Timestamp.fromHashAttributes(attrs)

            self.assertEqual(tm.getTid(), 2 ** 40)
            self.assertEqual(tm.getSeconds(), 2 ** 41)
            self.assertEqual(tm.getFractionalSeconds(), 2 ** 42)
        except Exception as e:
            self.fail(" testing conversion from uint64 attributes: " + str(e))

        try:
            # Test time attributes with a mix of uint32 and small uint64 values
            h64 = Hash()
            h64.setAs("tid", 2 ** 31, Types.UINT32)
            h64.setAs("sec", 2, Types.UINT64)
            h64.setAs("frac", 3, Types.UINT64)

            h = Hash()
            h["timestamp64"] = True
            h.setAttribute("timestamp64", "tid", h64["tid"])
            h.setAttribute("timestamp64", "sec", h64["sec"])
            h.setAttribute("timestamp64", "frac", h64["frac"])
            attrs = h.getAttributes("timestamp64")
            self.assertEqual(
                Timestamp.hashAttributesContainTimeInformation(attrs), True)
            tm = Timestamp.fromHashAttributes(attrs)

            self.assertEqual(tm.getTid(), 2 ** 31)
            self.assertEqual(tm.getSeconds(), 2)
            self.assertEqual(tm.getFractionalSeconds(), 3)
        except Exception as e:
            self.fail(
                " testing conversion from large uint32 "
                "/ small uint64 attributes: " + str(e))

        try:
            # Test failures in case of negative or string literals
            # as time attributes
            h = Hash()
            h["timestamp"] = True
            h.setAttribute("timestamp", "tid", -1)
            h.setAttribute("timestamp", "sec", "123")
            h.setAttribute("timestamp", "frac", 456)
            attrs = h.getAttributes("timestamp")
            self.assertEqual(
                Timestamp.hashAttributesContainTimeInformation(attrs), True)

            with self.assertRaises(Exception):
                Timestamp.fromHashAttributes(attrs)

            h.setAttribute("timestamp", "sec", 123)

            with self.assertRaises(Exception):
                Timestamp.fromHashAttributes(attrs)

            h.setAttribute("timestamp", "tid", 1)
            tm = Timestamp.fromHashAttributes(h.getAttributes("timestamp"))
            self.assertEqual(tm.getTid(), 1)
            self.assertEqual(tm.getSeconds(), 123)
            self.assertEqual(tm.getFractionalSeconds(), 456)

        except Exception as e:
            self.fail(
                " testing conversion from negative or ill-formed "
                "time attributes: " + str(e))


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import Hash, Logger


def test_logging_config_rules():
    # Python API: 'Logger.logXXX(message, category)
    # We are chatty in this test
    # But the idea is to only see OKs and never ERROR
    # There is no ASSERT unfortunately, so this test needs visual inspection
    config = Hash("level", "DEBUG")
    Logger.configure(config)
    Logger.logDebug("ERROR")

    Logger.useConsole()
    Logger.logDebug("OK")
    Logger.logDebug("OK", "a1")
    Logger.logDebug("OK", "a1.a2")
    Logger.logInfo("OK")
    Logger.logInfo("OK", "a1")
    Logger.logInfo("OK", "a1.a2")

    Logger.reset()
    Logger.logDebug("ERROR")
    Logger.logDebug("ERROR", "a1")
    Logger.logDebug("ERROR", "a1.a2")
    Logger.logInfo("ERROR")
    Logger.logInfo("ERROR", "a1")
    Logger.logInfo("ERROR", "a1.a2")

    Logger.useConsole("a1")
    Logger.logDebug("ERROR")
    Logger.logDebug("OK", "a1")
    Logger.logDebug("OK", "a1.a2")
    Logger.logInfo("ERROR")
    Logger.logInfo("OK", "a1")
    Logger.logInfo("OK", "a1.a2")

    Logger.setLevel("INFO")
    Logger.logDebug("ERROR")
    Logger.logDebug("ERROR", "a1")
    Logger.logDebug("ERROR", "a1.a2")
    Logger.logInfo("ERROR")
    Logger.logInfo("OK", "a1")
    Logger.logInfo("OK", "a1.a2")

    Logger.setLevel("WARN")
    Logger.logDebug("ERROR")
    Logger.logDebug("ERROR", "a1")
    Logger.logDebug("ERROR", "a1.a2")
    Logger.logInfo("ERROR")
    Logger.logInfo("ERROR", "a1")
    Logger.logInfo("ERROR", "a1.a2")


def test_logging_config_more():
    Logger.reset()
    config = Hash("level", "INFO")
    Logger.configure(config)
    Logger.useConsole()
    Logger.useFile("a1", False)  # do not inherit appenders from parents
    Logger.logDebug("ERROR")
    Logger.logDebug("ERROR", "a1")
    Logger.logDebug("ERROR", "a1.a2")
    Logger.logInfo("CONSOLE-OK")
    Logger.logInfo("FILE-OK", "a1")
    Logger.logInfo("FILE-OK", "a1.a2")


def test_logging_last_message():
    # Reset (remove all appenders) and clear the cache
    Logger.reset()

    content = Logger.getCachedContent(10)
    assert len(content) == 0

    # Setup the Logger
    maxmsgs = 20
    config = Hash("level", "INFO", "cache.maxNumMessages", maxmsgs)
    Logger.configure(config)
    Logger.useCache()

    # calling Logger.getCachedContent before logging will get an empty vector
    content = Logger.getCachedContent(10)
    assert len(content) == 0

    # log something
    for i in range(100):
        Logger.logDebug(f"This should not be logged - {i}", "VERBOSE_STUFF")
        Logger.logInfo(f"line - {i}", "INFORMATIVE_STUFF")

    # get the last 10 entries
    content = Logger.getCachedContent(10)
    assert len(content) == 10

    index = 90
    for entry in content:
        assert entry.has("timestamp") is True
        assert entry.get("category") == "INFORMATIVE_STUFF"
        assert entry.get("type") == "INFO"
        expected = f'line - {index}'
        index += 1
        assert expected == entry.get("message")

    # One can request more than the max config["cache.maxNumMessages"]
    # but will not get more than that
    content = Logger.getCachedContent(200)
    assert maxmsgs == len(content)
    index = 100 - maxmsgs
    for entry in content:
        # check that the timestamp is in, but do not check
        assert entry.has("timestamp") is True
        assert "INFORMATIVE_STUFF" == entry.get("category")
        assert "INFO" == entry.get("type")
        expected = f'line - {index}'
        assert expected == entry.get("message")
        index += 1



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import (
    ALARM_ELEMENT, BOOL_ELEMENT, FLOAT_ELEMENT, KARABO_CLASSINFO,
    KARABO_CONFIGURATION_BASE_CLASS, STATE_ELEMENT, STRING_ELEMENT,
    BinarySerializerHash, BinarySerializerSchema, Hash, MetricPrefix, Schema,
    TextSerializerHash, TextSerializerSchema, Unit, fullyEqual,
    loadHashFromFile, loadSchemaFromFile, saveHashToFile, saveSchemaToFile)
from karabo.common.alarm_conditions import AlarmCondition
from karabo.common.states import State


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("ShapeY", "1.0")
class ShapeY:
    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected)
            .key("shadowEnabled")
            .description("Shadow enabled")
            .displayedName("Shadow")
            .assignmentOptional()
            .defaultValue(False)
            .init()
            .commit(),
        )


@KARABO_CLASSINFO("CircleY", "1.0")
class CircleY(ShapeY):
    @staticmethod
    def expectedParameters(expected):
        (
            FLOAT_ELEMENT(expected)
            .key("radius")
            .alias(1)
            .description("The radius of the circle")
            .displayedName("Radius")
            .minExc(0)
            .maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(10)
            .init()
            .commit(),

            STATE_ELEMENT(expected).key("state")
            .commit(),

            STATE_ELEMENT(expected).key("stateN")
            .initialValue(State.NORMAL)
            .commit(),

            STATE_ELEMENT(expected).key("stateE")
            .initialValue(State.ERROR)
            .commit(),

            ALARM_ELEMENT(expected).key("alarm")
            .commit(),

            ALARM_ELEMENT(expected).key("alarmW")
            .initialValue(AlarmCondition.WARN)
            .commit(),

            ALARM_ELEMENT(expected).key("alarmA")
            .initialValue(AlarmCondition.ALARM)
            .commit(),

            STRING_ELEMENT(expected).key("status")
            .readOnly()
            .commit(),
        )


def test_serializer_Hash():
    config = Hash("indentation", -1)
    ser = TextSerializerHash.create("Xml", config)
    data = Hash('a.b.c', 1, 'x.y.z', [1, 2, 3, 4, 5, 6, 7])
    archive = ser.save(data)
    assert archive == '<?xml version="1.0"?><root KRB_Artificial="" \
KRB_Type="HASH"><a KRB_Type="HASH"><b KRB_Type="HASH"><c KRB_Type\
="INT32">1</c></b></a><x KRB_Type="HASH"><y KRB_Type="HASH"><z \
KRB_Type="VECTOR_INT32">1,2,3,4,5,6,7</z></y></x></root>'
    assert isinstance(archive, str) is True
    g = ser.load(archive)
    assert isinstance(g, Hash) is True
    assert fullyEqual(g, data) is True
    ser = BinarySerializerHash.create("Bin")
    archive = ser.save(data)
    assert isinstance(archive, bytes) is True
    g = ser.load(archive)
    assert isinstance(g, Hash) is True
    assert fullyEqual(g, data) is True


def test_serializer_schema():
    schema = Schema()
    CircleY.expectedParameters(schema)
    config = Hash("indentation", -1)
    ser = TextSerializerSchema.create("Xml", config)
    archive = ser.save(schema)
    assert isinstance(archive, str) is True
    schema2 = ser.load(archive)
    assert isinstance(schema2, Schema) is True
    assert schema.getParameterHash() == schema2.getParameterHash()
    ser = BinarySerializerSchema.create("Bin")
    archive = ser.save(schema)
    assert isinstance(archive, bytes) is True
    schema2 = ser.load(archive)
    assert isinstance(schema2, Schema) is True
    assert schema2.getParameterHash() == schema.getParameterHash()


def test_file_tools_hash():
    h = Hash("a", 10, "b.c", "Hallo World", "x.y.z", [1, 2, 3, 4, 5])
    saveHashToFile(h, '/tmp/hallo.xml')
    g = loadHashFromFile('/tmp/hallo.xml')
    assert fullyEqual(h, g) is True
    saveHashToFile(h, '/tmp/hallo.bin')
    g = loadHashFromFile('/tmp/hallo.bin')
    assert fullyEqual(h, g) is True


def test_file_tools_schema():
    schema = Schema()
    CircleY.expectedParameters(schema)
    saveSchemaToFile(schema, '/tmp/circle.xml')
    schema2 = loadSchemaFromFile('/tmp/circle.xml')
    assert fullyEqual(schema.getParameterHash(), schema2.getParameterHash())
    saveSchemaToFile(schema, '/tmp/circle.bin')
    schema2 = loadSchemaFromFile('/tmp/circle.bin')
    assert fullyEqual(schema.getParameterHash(), schema2.getParameterHash())



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from threading import Condition, Lock, Thread
from unittest import TestCase, main

from karabind import EventLoop, SignalSlotable


class Tests(TestCase):
    senderId = "sigSlotSender"
    receiverId = "sigSlotReceiver"

    received_value = None
    event_loop = None
    sigSender = None
    sigReceiver = None

    @classmethod
    def setUpClass(cls):
        # Start the EventLoop so that DeviceClient works properly
        cls.event_loop = Thread(target=EventLoop.work)
        cls.event_loop.daemon = True
        cls.event_loop.start()

        # Set up slots.
        cls.sigSender = SignalSlotable(cls.senderId)
        cls.sigSender.start()
        cls.sigReceiver = SignalSlotable(cls.receiverId)
        cls.sigReceiver.start()

    @classmethod
    def tearDownClass(cls):
        del cls.sigSender, cls.sigReceiver
        EventLoop.stop()
        cls.event_loop.join()

    def _return_value(self, val):
        """Return the received value (through _receive_value variable)."""

        with self.check_value_condition:
            self.received_value = val
            self.check_value_condition.notify()

    def call_and_compare(self, sent_value, timeout=3):
        """Call the _return_value function and compare the
        ``returned value`` to ``send value``.

        :param sent_value: a value to be sent
        :param timeout: How much time to wait for the callback to be called.
        """
        # Wait for the callback to be called.
        with self.check_value_condition:
            # Reset received_value.
            self.received_value = None
            self.sigReceiver.call(Tests.receiverId, "_return_value",
                                  sent_value)
            self.assertTrue(self.check_value_condition.wait(timeout),
                            "The callback was not called in time.")

            self.assertEqual(sent_value, self.received_value,
                             "The received value was not expected.")

    def test_type_handle(self):
        """Test if types are handled properly when when a socket is called."""

        # Register _return_value slot.

        self.received_value = None
        self.check_value_condition = Condition(Lock())
        self.sigReceiver.registerSlot(self._return_value)

        # Test how boolean values are handled.
        self.call_and_compare(True)
        self.call_and_compare(False)

        # Test how integers are handled.
        self.call_and_compare(2147483648)
        self.call_and_compare(4294967295)
        self.call_and_compare(0x89213A892)

        # Test how floating points are handled.
        self.call_and_compare(1.7976931348623157e+180)
        self.call_and_compare(float('inf'))

        # Test how strings are handled.
        self.call_and_compare(
            "This are not the droids that you are looking for.")
        # Test if null character is handled properly.
        self.call_and_compare(
            "This are not the droids that you are " + '\0' + "looking for.")
        # Test if a subset of Cyrillic characters is handled properly.
        self.call_and_compare(
            "М Н О П Р С Т У Ф Х Ц Ч Ш Щ Ъ Ы Ь Э Ю Я а б в г д")
        # Test if a subset of Latin-1 Supplement characters is
        # handled properly.
        self.call_and_compare("Ô Õ Ö × Ø Ù Ú Û Ü Ý Þ ß à á â ã ä å æ ç")

        # Test how a random user defined types are handled.
        class SimpleType:
            pass

        with self.assertRaises(TypeError):
            self.call_and_compare(SimpleType())
            # XXX: Additional tests need to be added (e.g. test Hash, ...).


if __name__ == "__main__":
    main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import unittest

from karabo.bound import (
    BinarySerializerHash, Hash, InputHash, OutputHash, TextSerializerHash,
    TextSerializerSchema, loadFromFile, saveToFile, similar)

from .configuration_example_classes import Shape


class Serializer_TestCase(unittest.TestCase):
    def test_textserializer_hash(self):
        try:
            config = Hash("indentation", -1)
            # create default serializer
            ser = TextSerializerHash.create("Xml", config)
            hash = Hash('a.b.c', 1, 'x.y.z', [1, 2, 3, 4, 5, 6, 7])
            archive = ser.save(hash)
            self.assertEqual(archive,
                             '<?xml version="1.0"?><root KRB_Artificial="" KRB_Type="HASH"><a KRB_Type="HASH"><b KRB_Type="HASH"><c KRB_Type="INT32">1</c></b></a><x KRB_Type="HASH"><y KRB_Type="HASH"><z KRB_Type="VECTOR_INT32">1,2,3,4,5,6,7</z></y></x></root>')  # noqa
            hash2 = ser.load(archive)
            self.assertEqual(similar(hash, hash2), True)

        except Exception as e:
            self.fail("test_textserializer_hash exception group 1: " + str(e))

    def test_textserializer_schema_xml(self):
        try:
            config = Hash("indentation", -1)
            ser = TextSerializerSchema.create("Xml", config)
            schema = Shape.getSchema("EditableCircle")
            archive = ser.save(schema)
            schema2 = ser.load(archive)
            self.assertEqual(similar(schema.getParameterHash(),
                                     schema2.getParameterHash()), True)

        except Exception as e:
            self.fail("test_textserializer_schema_xml exception group 1: "
                      + str(e))

    def test_binaryserializer_hash(self):
        try:
            ser = BinarySerializerHash.create("Bin")
            hash = Hash('a.b.c', 1, 'x.y.z', [1, 2, 3, 4, 5, 6, 7],
                        'e.f.e', None, 'e.g.h', [None, None, None, None, None])
            hash.setAttribute('e.g.h', 'attr1', None)
            hash.setAttribute('e.f.e', 'attrList', [None, None])
            archive = ser.save(hash)  # serialize hash
            self.assertEqual(archive.__class__.__name__, 'bytes')
            hash2 = ser.load(archive)
            self.assertTrue(similar(hash, hash2))

        except Exception as e:
            self.fail("test_binaryserializer_hash exception group 1: "
                      + str(e))

    def test_io_write_read_hash(self):
        try:
            config = Hash("filename", "/tmp/test_io_writeread_hash.xml",
                          "format", "Xml", "Xml.indentation", -1)
            out = OutputHash.create("TextFile", config)
            hash = Hash('a.b.c', 1, 'x.y.z', [1, 2, 3, 4, 5, 6, 7])
            out.write(hash)

            hash2 = Hash()
            config = Hash("filename", "/tmp/test_io_writeread_hash.xml")
            input = InputHash.create("TextFile", config)
            input.read(hash2)
            self.assertTrue(similar(hash, hash2))

        except Exception as e:
            self.fail("test_io_saveload_hash exception group 1: " + str(e))

    def test_io_savetofile_loadfromfile(self):
        try:
            h = Hash("a", 10, "b", "Hallo World")
            saveToFile(h, "/tmp/MyFile.xml", Hash("format", "Xml",
                                                  "Xml.indentation", 3))

            h2 = Hash()
            loadFromFile(h2, "/tmp/MyFile.xml")
            self.assertEqual(h2['a'], 10)
            self.assertEqual(h2['b'], "Hallo World")

        except Exception as e:
            self.fail("test_io_savetofile_loadfromfile exception group 1: "
                      + str(e))

        try:
            h = Hash("a", 10, "b", "Hallo World")
            saveToFile(h, "/tmp/MyFile.xml", Hash("format", "Xml",
                                                  "Xml.indentation", 1))

            h2 = loadFromFile("/tmp/MyFile.xml")
            self.assertEqual(h2['a'], 10)
            self.assertEqual(h2['b'], "Hallo World")

        except Exception as e:
            self.fail("test_io_savetofile_loadfromfile exception group 2: "
                      + str(e))


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import pytest

from karabo.bound import Hash, HashAttributes, HashAttributesNode, Types


def test_hashAttributes():
    # Test "has", "__contains__", "isType", "getType", "erase",
    #      "__delitem__", "size", "__len__", "empty", "bool",
    #      "clear",  "getNode", "get", "__getitem__", "getAs", "set",
    #      "__setitem__", "__iter__"

    h = Hash("a.b.c", "value")
    attrs = h.getAttributes("a.b.c")
    assert isinstance(attrs, HashAttributes)

    # Test 'empty'
    assert attrs.empty()

    # Test 'bool'
    assert not attrs

    # Test 'has' ...
    assert not attrs.has('a.b.c')

    # Test 'set'...
    attrs.set("aq", 12345678993758)
    assert attrs.has("aq")

    # Test 'get' ...
    assert attrs.get("aq") == 12345678993758

    # Test "__len__"
    assert len(attrs) == 1
    # add the second attribute ...
    attrs.set("ua", "interesting")

    # Test "__getitem__"
    assert attrs["ua"] == "interesting"
    assert len(attrs) == 2

    # Test "isType"
    # assert attrs.isType("ua", "STRING") is True
    # assert attrs.isType("ua", Types.STRING) is True

    # Test "size"
    assert attrs.size() == 2

    # add one more attribute ...
    attrs['minerale'] = [True, False]
    assert attrs["minerale"] == [True, False]

    # Test "getType", "isType"
    # check attribute types ...
    assert attrs.getType('minerale') == Types.VECTOR_BOOL
    # assert attrs.isType('minerale', "VECTOR_BOOL") is True

    # Test "__contains__ ..."
    assert "aq" in attrs

    # Test "erase" ...
    attrs.erase("aq")
    assert "aq" not in attrs

    # Test "getNode"
    node = attrs.getNode("minerale")
    assert isinstance(node, HashAttributesNode)
    # Tests for node can be found in "test_hash_attributes_node.py".

    # Test "__delitem__" ... reference to "minerale" node
    del node
    assert "minerale" in attrs
    # delete attribute
    del attrs['minerale']
    assert 'minerale' not in attrs
    # attribute 'ua' is still in attrs container
    assert not attrs.empty()

    # Test "clear"
    attrs.clear()
    assert attrs.empty()
    assert not bool(attrs)

    # set new attributes ...
    # Test 'set'
    attrs.set("a1", 12)
    attrs.set("a2", True)

    # Test "__setitem__" ...
    attrs["a3"] = 120_333_666
    attrs["a4"] = 3.141592
    attrs["a5"] = [1, 2, 3]
    attrs.set("a6", 'abrakadabra')
    attrs["a7"] = ['abra', 'kadabra']
    assert len(attrs) == 7
    assert attrs.size() == 7

    # Test "__iter__"
    it = iter(attrs)
    # Test first item
    n = next(it)
    assert isinstance(n, HashAttributesNode)
    assert n.getKey() == 'a1'
    assert n.getValue() == 12
    assert n.getType() == Types.INT32
    n = next(it)
    assert n.getKey() == 'a2'
    n = next(it)
    assert n.getKey() == 'a3'
    n = next(it)
    assert n.getKey() == 'a4'
    n = next(it)
    assert n.getKey() == 'a5'
    n = next(it)
    assert n.getKey() == 'a6'
    assert n.getValue() == 'abrakadabra'
    assert n.getType() == Types.STRING
    n = next(it)
    assert n.getKey() == 'a7'
    assert n.getValue() == ['abra', 'kadabra']
    assert n.getType() == Types.VECTOR_STRING
    # Test end of iteration: "StopIteration"
    with pytest.raises(StopIteration):
        n = next(it)

    attrs = h.getAttributes("a.b.c")

    # We do not need to test `getAs` here for each possible source and
    # target type since `getAs` uses the same conversion method as for
    # 'getValueAs'
    #
    # Test "getAs"
    assert attrs.getAs("a2", "BOOL") is True
    assert attrs.getAs("a2", Types.VECTOR_BOOL) == [True]
    assert attrs.getAs("a2", "VECTOR_UINT64") == [1]
    assert attrs.getAs("a2", "FLOAT") == 1.0
    assert attrs.getAs("a2", "COMPLEX_FLOAT") == (1+0j)
    assert attrs.getAs("a2", "VECTOR_COMPLEX_DOUBLE") == [(1+0j)]

    assert attrs.getAs("a1", Types.UINT32) == 12
    assert attrs.getAs("a1", "STRING") == '12'



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.bound import Hash, Types


def test_getAs():
    h = Hash("a", True)
    assert h.getAs("a", Types.STRING) == "1"
    assert h.getAs("a", Types.INT32) == 1
    assert h.getAs("a", Types.INT64) == 1
    assert h.getAs("a", Types.FLOAT) == 1.0
    assert h.getAs("a", Types.DOUBLE) == 1.0

    h = Hash("a", True)
    h.setAttribute("a", "a", True)
    assert h.getAttributeAs("a", "a", Types.STRING) == "1"
    assert h.getAttributeAs("a", "a", Types.INT32) == 1
    assert h.getAttributeAs("a", "a", Types.DOUBLE) == 1.0
    h.setAttribute("a", "b", 12)
    h.setAttribute("a", "c", 1.23)
    attrs = h.getAttributes("a")
    g = Hash("Z.a.b.c", "value")
    g.setAttributes("Z.a.b.c", attrs)
    assert g.getAttributeAs("Z.a.b.c", "a", Types.STRING) == "1"
    assert g.getAttributeAs("Z.a.b.c", "a", Types.INT32) == 1
    assert g.getAttributeAs("Z.a.b.c", "a", Types.DOUBLE) == 1.0

    # value is python list of boolean -> std::vector<bool>
    h = Hash("a", [False, False, False, False])
    assert h.getAs("a", Types.STRING) == "0,0,0,0"
    assert h.getAs("a", Types.VECTOR_INT32)[3] == 0

    # value is python list -> std::vector<char>
    h = Hash("a", ['1', '2', '3', '4'])
    assert h.getAs("a", Types.STRING) == "1,2,3,4"
    assert h.getAs("a", Types.VECTOR_INT32)[3] == 4

    h = Hash("a", [13, 13, 13, 13])
    assert h.getAs("a", Types.STRING) == "13,13,13,13"

    h = Hash("a", [-42])
    assert h.getAs("a", Types.STRING) == "-42"

    h = Hash("a", -2147483647)
    assert h.getAs("a", Types.STRING) == "-2147483647"

    h = Hash("a", 1234567890123456789)
    assert h.getAs("a", Types.STRING) == "1234567890123456789"
    assert h.getType("a") == Types.INT64
    assert str(h.getType("a")) == "INT64"

    h = Hash("a", 0.123456789123456)
    assert h.getAs("a", Types.STRING) == "0.123456789123456"
    assert h.getType("a") == Types.DOUBLE



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.bound import Hash


def test_empty():
    # Check properties of empty Hash
    h = Hash()
    assert len(h) == 0
    assert h.empty()


def test_one_property():
    # Check Hash with one property
    h = Hash('a', 1)
    assert len(h) == 1
    assert not h.empty()
    assert h.get('a') == 1
    assert h['a'] == 1


def test_two_properties():
    # Check Hash with 2 properties
    h = Hash('a', 1, 'b', 2.0)
    assert not h.empty()
    assert len(h) == 2
    assert h['a'] == 1
    assert h['b'] == 2.0


def test_six_properties():
    # Check Hash with 6 properties of different types
    h = Hash("a.b.c", 1, "b.c", 2.0, "c", 3.7, "d.e", "4",
             "e.f.g.h", [5, 5, 5, 5, 5], "F.f.f.f.f", Hash("x.y.z", 99))
    assert not h.empty()
    assert len(h) == 6
    assert h['a.b.c'] == 1
    assert h['b.c'] == 2.0
    assert h['c'] == 3.7
    assert h['d.e'] == "4"
    assert h['e.f.g.h'][0] == 5
    assert len(h['e.f.g.h']) == 5
    assert h['F.f.f.f.f']['x.y.z'] == 99
    assert h['F.f.f.f.f.x.y.z'] == 99
    # Make Hash flat
    flat = Hash()
    Hash.flatten(h, flat)
    assert not flat.empty()
    assert len(flat) == 6
    assert flat.get('a.b.c', ' ') == 1
    assert flat.get('b.c', ' ') == 2.0
    assert flat.get('c', ' ') == 3.7
    assert flat.get('d.e', ' ') == "4"
    assert flat.get('e.f.g.h', ' ')[0] == 5
    assert len(flat.get('e.f.g.h', ' ')) == 5
    assert flat.get('F.f.f.f.f.x.y.z', ' ') == 99

    # Make flat Hash unflatten again
    tree = Hash()
    flat.unflatten(tree)
    assert not tree.empty()
    assert len(tree) == 6
    assert tree['a.b.c'] == 1
    assert tree['b.c'] == 2.0
    assert tree['c'] == 3.7
    assert tree['d.e'] == "4"
    assert tree['e.f.g.h'][0] == 5
    assert len(tree['e.f.g.h']) == 5
    assert tree['F.f.f.f.f']['x.y.z'] == 99
    assert tree['F.f.f.f.f.x.y.z'] == 99



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
import socket
import time
import uuid
from collections.abc import Iterable
from threading import Thread

import pytest

from karabo.bound import (
    DeviceClient, EventLoop, Hash, Logger, Schema, startDeviceServer,
    stopDeviceServer)

timeoutSec = 15  # test_zcore_device_client2.py::test_device_client_sync_api
Logger.configure(Hash())


def test_device_client_sync_api():
    # Run CPP event loop in background ...
    loopThread = Thread(target=EventLoop.work)
    loopThread.start()

    # Create DeviceClient ... generated clientId ...
    c = DeviceClient()
    # test getInstanceId
    clientId = c.getInstanceId()
    (host, klass, pid) = clientId.split('_', 3)
    assert host == socket.gethostname().split(".", 1)[0]  # split domain away
    assert klass == 'DeviceClient'
    assert pid == str(os.getpid())

    # Test getPropertyHistory
    # Expect TymeoutError since DataLogReader is not running.
    # To reduce waiting time, set new internal timeout
    cachedTimeout = c.getInternalTimeout()
    c.setInternalTimeout(50)

    with pytest.raises(RuntimeError):
        c.getPropertyHistory("data/gen/22", "int32Property", "2023-10-25",
                             "2023-11-01")

    # Test getConfigurationFromPast...
    with pytest.raises(RuntimeError):
        c.getConfigurationFromPast("data/gen/22", "2023-11-01")

    # Test listInitConfigurations...
    h = c.listInitConfigurations("data/gen/22", "config1")
    assert h['success'] is False
    assert 'timed out' in h['reason']

    # Test 'saveInitConfiguration'...
    (success, errmsg) = c.saveInitConfiguration(
        "config1", ["data/gen/22", "data/gen/33"])
    assert success is False
    assert 'timed out' in errmsg

    # Restore default ...
    c.setInternalTimeout(cachedTimeout)

    # avoid id clash between test runs
    serverId = f'cppServer/{Hash.__module__}'
    deviceId = f'data/prop/{Hash.__module__}'

    # test registerInstance{New,Update}Monitor...
    instanceNewArg = instanceUpdatedArg = None

    def onInstanceNew(topologyEntry):
        nonlocal instanceNewArg
        instanceNewArg = topologyEntry

    def onInstanceUpdated(topologyEntry):
        nonlocal instanceUpdatedArg
        instanceUpdatedArg = topologyEntry

    c.registerInstanceNewMonitor(onInstanceNew)
    c.registerInstanceUpdatedMonitor(onInstanceUpdated)
    c.enableInstanceTracking()

    # Configure and start device server ...
    logLevel = "FATAL"
    config = Hash("serverId", serverId, "log.level", logLevel)

    # Test getServers...
    startDeviceServer(config)

    max_tries = 10
    n = 0
    server_not_present = True

    while n < max_tries and server_not_present:
        time.sleep(0.1)
        server_not_present = serverId not in c.getServers()
        n += 0.1

    assert not server_not_present

    # Check that instanceNew handler was called properly
    # (If 'is Hash' fails sometimes, better check first in a loop with timeout.
    # But as the client knows the server, instanceNew should have been called.)
    assert type(instanceNewArg) is Hash
    assert instanceNewArg.has("server." + serverId)
    serverInfo = instanceNewArg["server"]
    assert serverInfo.getAttribute(serverId, "type") == "server"
    assert serverInfo.getAttribute(serverId, "lang") == "cpp"
    assert serverInfo.getAttribute(serverId, "host") == host
    assert serverInfo.getAttribute(serverId, "serverId") == serverId
    assert serverInfo.getAttribute(serverId, "log") == logLevel

    # This test fails because instanceUpdate is not sent anymore by the C++
    # server. Removed for now, eventually, it will have to be revived; see
    # https://git.xfel.eu/Karabo/Framework/-/issues/983
    #
    # # Check that instanceUpdated handler was called properly. Note that
    # # here we rely on the fact that C++ server sends instanceUpdate soon
    # # after instanceNew, extending the known deviceClasses. (Instead of
    # # sending  instanceNew after plugins are known. If that changes, find
    # # something else for testing instanceUpdateHandler...)
    # timeout = 10
    # while timeout > 0:
    #     if type(instanceUpdatedArg) is Hash:
    #         break
    #     timeout -= 0.1
    #     time.sleep(0.1)

    # assert type(instanceUpdatedArg) is Hash
    # assert instanceUpdatedArg.has("server." + serverId)
    # serverUpdate = instanceUpdatedArg["server"]
    # assert serverUpdate.getAttribute(serverId, "type") == "server"
    # klasses = serverUpdate.getAttribute(serverId, "deviceClasses")
    # assert "PropertyTest" in klasses

    # Test getSystemTopology ...
    topology = c.getSystemTopology()
    assert isinstance(topology, Hash) is True
    assert 'client' in topology
    assert 'server' in topology
    assert serverId in topology["server"]
    # Test getSystemInformation
    system_information = c.getSystemInformation()
    assert isinstance(system_information, Hash)
    assert 'client' in system_information
    assert 'server' in system_information

    # Test exists...
    a, b = c.exists(serverId)
    assert a is True
    # Test getClasses... (can test directly since waited already for update)
    classes = c.getClasses(serverId)
    assert isinstance(classes, list)
    assert "PropertyTest" in classes

    # Test getClassProperties...
    props = c.getClassProperties(serverId, "PropertyTest")
    assert isinstance(props, list) is True
    assert 'classId' in props
    assert 'deviceId' in props
    assert 'serverId' in props
    assert 'karaboVersion' in props

    # Test instantiate...
    instanceNewArg = None
    #     synopsis: instantiate(serverId, classId, config, timeout)
    c.instantiate(serverId, Hash('PropertyTest.deviceId', deviceId),
                  timeoutSec)
    # Test getDevices...
    assert deviceId in c.getDevices(serverId)
    assert type(instanceNewArg) is Hash
    assert instanceNewArg.has("device." + deviceId)
    deviceInfo = instanceNewArg["device"]
    assert deviceInfo.getAttribute(deviceId, "type") == "device"
    assert deviceInfo.getAttribute(deviceId, "host") == host
    assert deviceInfo.getAttribute(deviceId, "serverId") == serverId

    # Test get/set ...
    assert c.get(deviceId, 'node.counter') == 0
    c.set(deviceId, 'node.counter', 10)
    assert c.get(deviceId, 'node.counter') == 10
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # Schema update monitor
    onSchemaUpdatedArg1 = onSchemaUpdatedArg2 = None

    def onSchemaUpdated(devId, schema):
        nonlocal onSchemaUpdatedArg1, onSchemaUpdatedArg2
        onSchemaUpdatedArg1 = devId
        onSchemaUpdatedArg2 = schema

    c.registerSchemaUpdatedMonitor(onSchemaUpdated)

    # The following is needed to keep the schema updates coming for longer
    # than the next 15 seconds (C++ DeviceClient::CONNECTION_KEEP_ALIVE)
    c.registerDeviceForMonitoring(deviceId)

    c.execute(deviceId, "slotUpdateSchema")
    assert type(onSchemaUpdatedArg1) is str
    assert onSchemaUpdatedArg1 == deviceId
    assert type(onSchemaUpdatedArg2) is Schema

    # To see that 15 seconds after unregistration schema updates are not
    # received anymore, we would have to sleep that long which would delay
    # tests, so its is commented. Still we test that the method can be called.
    c.unregisterDeviceFromMonitoring(deviceId)
    # time.sleep(15)
    # onSchemaUpdatedArg1 = onSchemaUpdatedArg1 = None
    # c.execute(deviceId, "slotUpdateSchema")
    # assert onSchemaUpdatedArg1 is None  # handler not called

    onPropertyChangeArgs = None

    # Switch on monitor for device property
    def onPropertyChange(devId, key, value, ts):
        nonlocal onPropertyChangeArgs
        onPropertyChangeArgs = (devId, key, value, ts)

    c.registerPropertyMonitor(deviceId, 'int32Property', onPropertyChange)

    # test 'set' ...
    c.set(deviceId, 'int32Property', 48000000)
    assert type(onPropertyChangeArgs) is tuple
    assert onPropertyChangeArgs[0] == deviceId
    assert onPropertyChangeArgs[1] == "int32Property"
    assert onPropertyChangeArgs[2] == 48000000
    assert onPropertyChangeArgs[3].toIso8601Ext().endswith("Z")

    c.set(deviceId, 'int32Property', 777)

    c.unregisterPropertyMonitor(deviceId, 'int32Property')

    # Test execute...
    # increment node.counter ...
    for i in range(3):
        c.execute(deviceId, 'node.increment', 5)

    assert c.get(deviceId, 'node.counter') == 13
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # Test getDeviceSchema...
    device_schema = c.getDeviceSchema(deviceId)
    assert 'outputFrequency' in device_schema.getParameterHash()
    assert 'output.connections' in device_schema.getParameterHash()
    # Test getActiveSchema...
    active_schema = c.getActiveSchema(deviceId)
    assert 'int16Property' in active_schema.getParameterHash()
    assert 'table' in active_schema.getParameterHash()
    # Test getClassSchema...
    class_schema = c.getClassSchema(serverId, 'PropertyTest')
    assert 'karaboVersion' in class_schema.getParameterHash()
    assert 'boolProperty' in class_schema.getParameterHash()
    # Test getProperties...
    props = c.getProperties(deviceId)
    assert 'status' in props
    assert 'vectors.boolProperty' in props
    # Test getCurrentlyExecutableCommands...
    comms = c.getCurrentlyExecutableCommands(deviceId)
    assert 'writeOutput' in comms
    assert 'eosOutput' in comms
    # Test getCurrentlySettableProperties...
    props = c.getCurrentlySettableProperties(deviceId)
    assert 'int32Property' in props
    assert 'boolProperty' in props
    # Test getOutputChannelNames...
    names = c.getOutputChannelNames(deviceId)
    assert isinstance(names, Iterable)
    assert 'output' in names
    # Test getOutputChannelSchema...
    schema = c.getOutputChannelSchema(deviceId, 'output')
    assert 'node.ndarray.shape' in schema
    assert 'node.image.pixels.data' in schema
    assert 'node.image.pixels.shape' in schema

    # c.setDeviceMonitorInterval(5)

    # Switch on the monitoring for device properties...
    onDeviceParametersChangeArg1 = onDeviceParametersChangeArg2 = None

    def onDeviceParametersChange(devId, configChanged):
        nonlocal onDeviceParametersChangeArg1, onDeviceParametersChangeArg2
        onDeviceParametersChangeArg1 = devId
        onDeviceParametersChangeArg2 = configChanged

    c.registerDeviceMonitor(deviceId, onDeviceParametersChange)

    c.execute(deviceId, 'node.increment', 5)
    assert onDeviceParametersChangeArg1 == deviceId
    assert type(onDeviceParametersChangeArg2) is Hash
    assert onDeviceParametersChangeArg2.has("node.counter")

    # Switch off monitoring for device properties...
    c.unregisterDeviceMonitor(deviceId)

    for i in range(6, 8):
        c.execute(deviceId, 'node.increment', 5)

    # Test instanceGoneMonitor
    instanceGoneArg1 = instanceGoneArg2 = None

    def onInstanceGone(instId, info):
        nonlocal instanceGoneArg1, instanceGoneArg2
        instanceGoneArg1 = instId
        instanceGoneArg2 = info

    c.registerInstanceGoneMonitor(onInstanceGone)

    # shutdown device ...
    # Test killDevice and gone monitor...
    c.killDevice(deviceId, 10)
    assert type(instanceGoneArg1) is str
    assert instanceGoneArg1 == deviceId
    assert type(instanceGoneArg2) is Hash
    assert instanceGoneArg2["type"] == "device"
    assert instanceGoneArg2["classId"] == "PropertyTest"
    assert instanceGoneArg2["host"] == host
    assert instanceGoneArg2["status"] == "ok"

    # de-register device server...
    stopDeviceServer(serverId)
    Logger.reset()
    EventLoop.stop()
    loopThread.join()


def test_device_client_async_api():
    # Run CPP event loop in background ...
    loopThread = Thread(target=EventLoop.work)
    loopThread.start()

    # Start device server (suppress plugins scan) ...
    serverId = f'cppServer/{uuid.uuid4()}'
    config = Hash("serverId", serverId)
    startDeviceServer(config)
    # Create DeviceClient ...
    clientId = str(uuid.uuid4())
    c = DeviceClient(clientId)
    assert c.getInstanceId() == clientId

    # Test setInternalTimeout/getInternalTimeout
    assert c.getInternalTimeout() == 3000
    c.setInternalTimeout(1000)
    to = c.getInternalTimeout()
    assert to == 1000

    # Test instantiateNoWait ...
    #     synopsis: instantiateNoWait(serverId, config)
    deviceId = f'data/prop/{uuid.uuid4()}'
    a, b = c.exists(deviceId)
    assert a is False
    c.instantiateNoWait(serverId, Hash('PropertyTest.deviceId', deviceId))

    trials = 1000
    while trials > 0 and deviceId not in c.getDevices(serverId):
        time.sleep(0.002)
        trials -= 1
    a, b = c.exists(deviceId)
    assert a is True
    assert deviceId in c.getDevices(serverId)

    assert c.get(deviceId, 'node.counter') == 0

    c.setNoWait(deviceId, 'node.counter', 10)
    trials = 1000
    while trials > 0 and c.get(deviceId, 'node.counter') != 10:
        time.sleep(0.02)
        trials -= 1
    assert c.get(deviceId, 'node.counter') == 10
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # increment counter
    for i in range(3):
        c.executeNoWait(deviceId, 'node.increment')
    trials = 1000
    while trials > 0 and c.get(deviceId, 'node.counter') != 13:
        time.sleep(0.02)
        trials -= 1

    assert c.get(deviceId, 'node.counter') == 13
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # getDeviceSchemaNoWait (pre-cache) ...
    trials = 1000
    while trials > 0 and c.getDeviceSchemaNoWait(deviceId).empty():
        time.sleep(0.02)
        trials -= 1
    schema = c.getDeviceSchemaNoWait(deviceId)
    assert schema.empty() is False

    # shutdown device ...
    c.killDeviceNoWait(deviceId)
    trials = 1000
    while trials > 0 and deviceId in c.getDevices(serverId):
        time.sleep(0.002)
        trials -= 1
    assert deviceId not in c.getDevices(serverId)

    # stop and de-register device server...
    stopDeviceServer(serverId)
    Logger.reset()
    EventLoop.stop()
    loopThread.join()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.bound import Hash, similar


def test_merge():
    h1 = Hash("a", 1,
              "b", 2,
              "c.b[0].g", 3,
              "c.c[0].d", 4,
              "c.c[1]", Hash("a.b.c", 6),
              "d.e", 7)

    h2 = Hash("a", 21,
              "b.c", 22,
              "c.b[0]", Hash("key", "value"),
              "c.b[1].d", 24,
              "e", 27)

    h1 += h2
    assert h1.has("a")
    assert h1.get("a") == 21
    assert h1["a"] == 21
    assert h1.has("b")
    assert not h1.has("c.b.d")
    assert h1.has("c.b[0]")
    assert h1.has("c.b[1]")
    assert not h1.has("c.b[2]")
    assert h1.get("c.b[1].d") == 24
    assert h1.has("c.c[0].d")
    assert h1.has("c.c[1].a.b.c")
    assert h1.has("d.e")
    assert h1.has("e")

    h3 = h1
    assert similar(h1, h3)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.bound import Hash


def get_hash():
    return Hash("should", 1, "be", 2, "iterated", 3, "in", 4,
                "correct", 5, "order", 6)


def test_iteration_1():
    insertionOrder = []
    h = get_hash()
    for k in h:
        insertionOrder.append(str(k))
    assert insertionOrder[0] == "should"
    assert insertionOrder[1] == "be"
    assert insertionOrder[2] == "iterated"
    assert insertionOrder[3] == "in"
    assert insertionOrder[4] == "correct"
    assert insertionOrder[5] == "order"


def test_iteration_2():
    alphaNumericOrder = []
    h = get_hash()
    for k in h:
        alphaNumericOrder.append(k.getKey())
    alphaNumericOrder.sort()
    assert alphaNumericOrder[0] == "be"
    assert alphaNumericOrder[1] == "correct"
    assert alphaNumericOrder[2] == "in"
    assert alphaNumericOrder[3] == "iterated"
    assert alphaNumericOrder[4] == "order"
    assert alphaNumericOrder[5] == "should"


def test_iteration_3():
    h = Hash(get_hash())
    h.set("be", "2")  # Has no effect on order

    insertionOrder = []
    for k in h:
        insertionOrder.append(str(k.getKey()))

    assert insertionOrder[0] == "should"
    assert insertionOrder[1] == "be"
    assert insertionOrder[2] == "iterated"
    assert insertionOrder[3] == "in"
    assert insertionOrder[4] == "correct"
    assert insertionOrder[5] == "order"


def test_iteration_4():
    alphaNumericOrder = []
    h = get_hash()
    for k in h:
        alphaNumericOrder.append(str(k.getKey()))
    alphaNumericOrder.sort()
    assert alphaNumericOrder[0] == "be"
    assert alphaNumericOrder[1] == "correct"
    assert alphaNumericOrder[2] == "in"
    assert alphaNumericOrder[3] == "iterated"
    assert alphaNumericOrder[4] == "order"
    assert alphaNumericOrder[5] == "should"


def test_iteration_5():
    h = Hash(get_hash())
    h.erase("be")
    # Must be last element in sequence now
    h.set("be", "2")

    insertionOrder = []
    for k in h:
        insertionOrder.append(str(k.getKey()))

    assert insertionOrder[0] == "should"
    assert insertionOrder[1] == "iterated"
    assert insertionOrder[2] == "in"
    assert insertionOrder[3] == "correct"
    assert insertionOrder[4] == "order"
    assert insertionOrder[5] == "be"


def test_iteration_6():
    alphaNumericOrder = []
    h = get_hash()
    for k in h:
        alphaNumericOrder.append(str(k.getKey()))
    alphaNumericOrder.sort()

    assert alphaNumericOrder[0] == "be"
    assert alphaNumericOrder[1] == "correct"
    assert alphaNumericOrder[2] == "in"
    assert alphaNumericOrder[3] == "iterated"
    assert alphaNumericOrder[4] == "order"
    assert alphaNumericOrder[5] == "should"


def test_iteration_7():
    tmp = []  # create empty set
    h = get_hash()
    h.getKeys(tmp)  # fill set by keys
    tmp.sort()
    it = iter(tmp)

    assert str(next(it)) == "be"
    assert str(next(it)) == "correct"
    assert str(next(it)) == "in"
    assert str(next(it)) == "iterated"
    assert str(next(it)) == "order"
    assert str(next(it)) == "should"


def test_iteration_8():
    tmp = []  # create empty vector
    h = get_hash()
    h.getKeys(tmp)  # fill vector by keys
    it = iter(tmp)

    assert str(next(it)) == "should"
    assert str(next(it)) == "be"
    assert str(next(it)) == "iterated"
    assert str(next(it)) == "in"
    assert str(next(it)) == "correct"
    assert str(next(it)) == "order"


def test_iteration_9():
    h = Hash("b", "bla-la-la", "a.b.c", 1, "abc.2", 2.2222222,
             "a.b.c1.d", "abc1d", "abc.1", 1.11111111)

    ll = []
    h.getKeys(ll)  # use top level keys

    assert len(ll) == 3
    i = iter(ll)  # "canonical" order: on every level insertion order
    assert str(next(i)) == "b"
    assert str(next(i)) == "a"
    assert str(next(i)) == "abc"

    ll = []
    h.getPaths(ll)   # use full keys

    assert len(ll) == 5
    i = iter(ll)
    assert str(next(i)) == "b"
    assert str(next(i)) == "a.b.c"
    assert str(next(i)) == "a.b.c1.d"
    assert str(next(i)) == "abc.2"
    assert str(next(i)) == "abc.1"
